# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the UCCGD ansatz class. It provides a chemically
inspired ansatz and is an implementation of the classical unitary CCGD ansatz.
Generalized double excitation determinants, in accordance
with the system number of electron and spin, are considered. For more
information about this ansatz, see references below.

Refs:

"""

import numpy as np
from openfermion import hermitian_conjugated
from openfermion import FermionOperator as ofFermionOperator
from itertools import combinations_with_replacement, product

from tangelo.linq import Circuit
from tangelo.toolboxes.ansatz_generator.ansatz import Ansatz
from tangelo.toolboxes.ansatz_generator.ansatz_utils import get_exponentiated_qubit_operator_circuit
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping
from tangelo.toolboxes.qubit_mappings.statevector_mapping import get_reference_circuit


class UCCGD(Ansatz):
    """This class implements the UCCGD ansatz. This implies that the
    mean-field is computed with the RHF or ROHF reference integrals.

    Args:
        molecule (SecondQuantizedMolecule) : The molecular system.
        mapping (str) : one of the supported qubit mapping identifiers. Default:
            "JW".
        up_then_down (bool): change basis ordering putting all spin up orbitals
            first, followed by all spin down. Default, False (i.e. has
            alternating spin up/down ordering).
    """

    def __init__(self, molecule, mapping="JW", up_then_down=False):

        self.n_spinorbitals = molecule.n_active_sos
        self.n_electrons = molecule.n_active_electrons
        self.spin = molecule.spin

        self.qubit_mapping = mapping
        self.up_then_down = up_then_down

        # Later: refactor to handle various flavors of UCCSD
        if self.n_spinorbitals % 2 != 0:
            raise ValueError("The total number of spin-orbitals should be even.")
        self.n_spatial_orbitals = self.n_spinorbitals // 2
        n_mos = self.n_spatial_orbitals

        self.n_var_params = len(list(combinations_with_replacement(range(n_mos), 4))) - n_mos

        # Supported reference state initialization
        # TODO: support for others
        self.supported_reference_state = {"HF", "zero"}
        # Supported var param initialization
        self.var_params_default = "ones"
        self.supported_initial_var_params = {"ones", "random"}

        # Default initial parameters for initialization
        self.default_reference_state = "HF"

        self.var_params = None
        self.circuit = None

    def set_var_params(self, var_params=None):
        """Set values for variational parameters, such as zeros, random numbers,
        providing some keywords for users, and also supporting direct user input
        (list or numpy array). Return the parameters so that workflows such as
        VQE can retrieve these values.
        """

        if var_params is None:
            var_params = self.var_params_default

        if isinstance(var_params, str):
            if (var_params not in self.supported_initial_var_params):
                raise ValueError(f"Supported keywords for initializing variational parameters: {self.supported_initial_var_params}")
            if var_params == "ones":
                initial_var_params = np.ones((self.n_var_params,), dtype=float)
            elif var_params == "random":
                initial_var_params = 1.e-2 * (np.random.random((self.n_var_params,)) - 0.5)
        else:
            initial_var_params = np.array(var_params)
            if initial_var_params.size != self.n_var_params:
                raise ValueError(f"Expected {self.n_var_params} variational parameters but "
                                 f"received {initial_var_params.size}.")
        self.var_params = initial_var_params
        return initial_var_params

    def prepare_reference_state(self):
        """Returns circuit preparing the reference state of the ansatz (e.g
        prepare reference wavefunction with HF, multi-reference state, etc).
        These preparations must be consistent with the transform used to obtain
        the qubit operator.
        """

        if self.default_reference_state not in self.supported_reference_state:
            raise ValueError(f"Only supported reference state methods are:{self.supported_reference_state}")

        if self.default_reference_state == "HF":
            return get_reference_circuit(n_spinorbitals=self.n_spinorbitals,
                                         n_electrons=self.n_electrons,
                                         mapping=self.qubit_mapping,
                                         up_then_down=self.up_then_down,
                                         spin=self.spin)
        if self.default_reference_state == "zero":
            return Circuit()

    def build_circuit(self, var_params=None):
        """Build and return the quantum circuit implementing the state
        preparation ansatz (with currently specified initial_state and
        var_params).
        """

        if var_params is not None:
            self.set_var_params(var_params)
        elif self.var_params is None:
            self.set_var_params()

        # Prepare reference state circuit
        reference_state_circuit = self.prepare_reference_state()

        qubit_op = self._get_qubit_operator()

        self.qu_op_dict = qubit_op.terms
        self.pauli_order = list(self.qu_op_dict.items())

        # Initialize dictionary of qubit_op terms for each UCCGD step
        self.pauli_to_angles_mapping = dict()

        # uccgd_circuit = trotterize(qubit_op, variational=True)
        uccgd_circuit = get_exponentiated_qubit_operator_circuit(qubit_op, variational=True, pauli_order=self.pauli_order)

        # skip over the reference state circuit if it is empty
        if reference_state_circuit.size != 0:
            self.circuit = reference_state_circuit + uccgd_circuit
        else:
            self.circuit = uccgd_circuit

    def update_var_params(self, var_params):
        """Shortcut: set value of variational parameters in the already-built
        ansatz circuit member. Preferable to rebuilt your circuit from scratch,
        which can be an involved process.
        """

        self.set_var_params(var_params)

        qubit_op = self._get_qubit_operator()
        qu_op_dict = qubit_op.terms

        if set(qu_op_dict) != set(self.qu_op_dict):
            self.build_circuit(var_params)
        else:
            for i, (term, _) in enumerate(self.pauli_order):
                value = 2*qu_op_dict[term] if qu_op_dict[term] >= 0. else 4*np.pi+2*qu_op_dict[term]
                self.circuit._variational_gates[i].parameter = value

    def _get_qubit_operator(self):
        """Construct UCCGD FermionOperator for variational
        parameters, and translate to QubitOperator via relevant qubit mapping.

        Returns:
            QubitOperator: qubit-encoded elements of the UCCGD
        """
        fermion_op = ofFermionOperator()
        n_mos = self.n_spinorbitals // 2
        p = -1
        for indices in combinations_with_replacement(range(n_mos), 4):
            if len(set(indices)) >= 2:  # are they not all equal
                u, w, v, t = indices
                p = p + 1
                for sig, tau in product(range(2), repeat=2):
                    c_op = ofFermionOperator(((2*t+sig, 1), (2*v+tau, 1), (2*w+tau, 0), (2*u+sig, 0)), self.var_params[p])
                    c_op += ofFermionOperator(((2*v+sig, 1), (2*t+tau, 1), (2*u+tau, 0), (2*w+sig, 0)), self.var_params[p])
                    fermion_op += c_op - hermitian_conjugated(c_op)

        qubit_op = fermion_to_qubit_mapping(fermion_operator=fermion_op,
                                            mapping=self.qubit_mapping,
                                            n_spinorbitals=self.n_spinorbitals,
                                            n_electrons=self.n_electrons,
                                            up_then_down=self.up_then_down)

        # Cast all coefs to floats (rotations angles are real)
        for key in qubit_op.terms:
            qubit_op.terms[key] = float(qubit_op.terms[key].imag)
        qubit_op.compress()
        return qubit_op



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    A test class to check that the simulator class functionalities are behaving as expected for the different
    backends and input possible.
"""

import unittest
import os
import time
import numpy as np
from openfermion.ops import QubitOperator
from openfermion import load_operator

from tangelo.linq import Gate, Circuit, translator, Simulator
from tangelo.linq.gate import PARAMETERIZED_GATES
from tangelo.helpers.utils import installed_simulator, installed_sv_simulator, installed_backends

path_data = os.path.dirname(os.path.abspath(__file__)) + '/data'

# Simple circuit for superposition, also tells us qubit ordering as well immediately from the statevector
# probabilities : |00> = 0.5  |01> = 0.5
circuit1 = Circuit([Gate("H", 0)], n_qubits=2)
# 2-qubit circuit checking all the basic gates that are not defined up to a convention (e.g unambiguous)
mygates = [Gate("H", 0), Gate("S", 0), Gate("X", 0), Gate("T", 1), Gate("Y", 1), Gate("Z", 1)]
mygates += [Gate("CNOT", 1, control=0)]
circuit2 = Circuit(mygates)
# Circuit for the parametrized rotation gates Rx and Ry. Some convention about the sign of theta or a phase may appear
circuit3 = Circuit([Gate("RX", 0, parameter=2.), Gate("RY", 1, parameter=-1.)])
# Circuit for the parametrized rotation gate Rz. Some convention about the sign of theta or a phase may appear
circuit4 = Circuit([Gate("RZ", 0, parameter=2.)], n_qubits=2)
# Circuit that tests all gates that are supported on all simulators
init_gates = [Gate('H', 0), Gate('X', 1), Gate('H', 2)]
one_qubit_gate_names = ["H", "X", "Y", "Z", "S", "T", "RX", "RY", "RZ", "PHASE"]
one_qubit_gates = [Gate(name, target=0) if name not in PARAMETERIZED_GATES else Gate(name, target=0, parameter=0.5)
                       for name in one_qubit_gate_names]
one_qubit_gates += [Gate(name, target=1) if name not in PARAMETERIZED_GATES else Gate(name, target=1, parameter=0.2)
                        for name in one_qubit_gate_names]
two_qubit_gate_names = ["CNOT", "CH", "CX", "CY", "CZ", "CRX", "CRY", "CRZ", "CPHASE"]
two_qubit_gates = [Gate(name, target=1, control=0) if name not in PARAMETERIZED_GATES
                       else Gate(name, target=1, control=0, parameter=0.5) for name in two_qubit_gate_names]
swap_gates = [Gate('SWAP', target=[1, 0]), Gate('CSWAP', target=[1, 2], control=0)]
circuit5 = Circuit(init_gates + one_qubit_gates + two_qubit_gates + swap_gates)
# Circuit preparing a mixed-state (e.g containing a MEASURE instruction in the middle of the circuit)
circuit_mixed = Circuit([Gate("RX", 0, parameter=2.), Gate("RY", 1, parameter=-1.), Gate("MEASURE", 0), Gate("X", 0)])

# Operators for testing the get_expectation_value functions
op1 = 1.0 * QubitOperator('Z0')  # all Z
op2 = 1.0 * QubitOperator('X1 Y0')  # X and Y
op3 = 1.0 * QubitOperator('Y0') - 2.0 * QubitOperator('Z0 X1')  # Linear combination
op4 = 1.0 * QubitOperator('Z0 Y1 X2')  # Operates on more qubits than circuit size

circuits = [circuit1, circuit2, circuit3, circuit4, circuit5]
ops = [op1, op2, op3]

# Reference results
ref_freqs = list()
ref_freqs.append({'00': 0.5, '10': 0.5})
ref_freqs.append({'01': 0.5, '10': 0.5})
ref_freqs.append({'00': 0.2248275934887, '10': 0.5453235594453, '01': 0.06709898823771, '11': 0.16274985882821})
ref_freqs.append({'00': 1.0})
ref_freqs.append({'000': 0.15972060437359714, '100': 0.2828171838599203, '010': 0.03984122195648572,
                  '110': 0.28281718385992016, '001': 0.15972060437359714, '101': 0.017620989809996816,
                  '011': 0.039841221956485706, '111': 0.01762098980999681})
reference_exp_values = np.array([[0., 0., 0.], [0., -1., 0.], [-0.41614684, 0.7651474, -1.6096484], [1., 0., 0.],
                                 [-0.20175269, -0.0600213, 1.2972912]])
reference_mixed = {'01': 0.163, '11': 0.066, '10': 0.225, '00': 0.545}  # With Qiskit noiseless, 1M shots


def assert_freq_dict_almost_equal(d1, d2, atol):
    """ Utility function to check whether two frequency dictionaries are almost equal, for arbitrary tolerance """
    if d1.keys() != d2.keys():
        raise AssertionError("Dictionary keys differ. Frequency dictionaries are not almost equal.\n"
                             f"d1 keys: {d1.keys()} \nd2 keys: {d2.keys()}")
    else:
        for k in d1.keys():
            if abs(d1[k] - d2[k]) > atol:
                raise AssertionError(f"Dictionary entries beyond tolerance {atol}: \n{d1} \n{d2}")
    return True


class TestSimulateAllBackends(unittest.TestCase):

    def test_get_exp_value_operator_too_long(self):
        """ Ensure an error is returned if the qubit operator acts on more qubits than are present in the circuit """
        for b in installed_simulator:
            simulator = Simulator(target=b, n_shots=1)
            self.assertRaises(ValueError, simulator.get_expectation_value, op4, circuit1)

    def test_get_exp_value_empty_operator(self):
        """ If qubit operator is empty, the expectation value is 0 and no computation occurs """
        for b in installed_simulator:
            simulator = Simulator(target=b, n_shots=1)
            exp_value = simulator.get_expectation_value(QubitOperator(), circuit1)
            self.assertTrue(exp_value == 0.)

    def test_get_exp_value_constant_operator(self):
        """ The expectation of the identity term must be 1. """
        for b in installed_simulator:
            simulator = Simulator(target=b, n_shots=1)
            const_op = QubitOperator()
            const_op.terms = {(): 777.}
            exp_value = simulator._get_expectation_value_from_frequencies(const_op, circuit1)
            self.assertTrue(exp_value == 777.)

    def test_simulate_mixed_state(self):
        """ Test mid-circuit measurement (mixed-state simulation) for compatible/testable formats and backends.
        Mixed-state do not have a statevector representation, as they are a statistical mixture of several statevectors.
        Simulating individual shots is suitable.

        Some simulators are NOT good at this, by design
        """

        results = dict()
        for b in installed_simulator:
            sim = Simulator(target=b, n_shots=10 ** 5)
            results[b], _ = sim.simulate(circuit_mixed)
            assert_freq_dict_almost_equal(results[b], reference_mixed, 1e-2)

    def test_get_exp_value_mixed_state(self):
        """ Test expectation value for mixed-state simulation. Computation done by drawing individual shots.
        Some simulators are NOT good at this, by design (ProjectQ). """

        reference = 0.41614683  # Exact value
        results = dict()
        for b in installed_simulator:
            sim = Simulator(target=b, n_shots=10 ** 5)
            results[b] = sim.get_expectation_value(op1, circuit_mixed)
            np.testing.assert_almost_equal(results[b], reference, decimal=2)


class TestSimulateStatevector(unittest.TestCase):

    def test_simulate_statevector(self):
        """ Must return correct frequencies for simulation of different quantum circuits with statevector """
        for b in installed_sv_simulator:
            simulator = Simulator(target=b)
            for i, circuit in enumerate(circuits):
                frequencies, _ = simulator.simulate(circuit)
                assert_freq_dict_almost_equal(ref_freqs[i], frequencies, atol=1e-5)

    def test_simulate_nshots_from_statevector(self):
        """
            Test the generation of samples following the distribution given by the exact frequencies obtained
            with a statevector simulator. For n_shots high enough, the resulting distribution must approximate
            the exact one.
        """
        for b in installed_sv_simulator:
            simulator = Simulator(target=b, n_shots=10**6)
            for i, circuit in enumerate(circuits):
                frequencies, _ = simulator.simulate(circuit)
                assert_freq_dict_almost_equal(ref_freqs[i], frequencies, atol=1e-2)

    def test_simulate_empty_circuit_from_statevector(self):
        """ Test the generation of frequencies using an initial_statevector and an empty_circuit """
        for b in installed_sv_simulator:
            simulator = Simulator(target=b)
            for i, circuit in enumerate(circuits):
                _, statevector = simulator.simulate(circuit, return_statevector=True)
                frequencies, _ = simulator.simulate(Circuit(n_qubits=circuit.width), initial_statevector=statevector)
                assert_freq_dict_almost_equal(ref_freqs[i], frequencies, atol=1e-5)

    def test_get_exp_value_from_statevector(self):
        """ Compute the expectation value from the statevector for each statevector backend """

        for b in installed_sv_simulator:
            simulator = Simulator(target=b)
            exp_values = np.zeros((len(circuits), len(ops)), dtype=float)
            for i, circuit in enumerate(circuits):
                for j, op in enumerate(ops):
                    exp_values[i][j] = simulator._get_expectation_value_from_statevector(op, circuit)
            np.testing.assert_almost_equal(exp_values, reference_exp_values, decimal=5)

    def test_get_exp_value_from_frequencies_using_initial_statevector(self):
        """ Test the method computing the expectation value from frequencies, with a given simulator
            by generating the statevector first and sampling using an empty state_prep_circuit
        """

        for b in installed_sv_simulator:
            simulator = Simulator(target=b)
            exp_values = np.zeros((len(circuits), len(ops)), dtype=float)
            for i, circuit in enumerate(circuits):
                _, statevector = simulator.simulate(circuit, return_statevector=True)
                for j, op in enumerate(ops):
                    exp_values[i][j] = simulator._get_expectation_value_from_frequencies(op,
                                                                                         Circuit(n_qubits=circuit.width),
                                                                                         initial_statevector=statevector)
            np.testing.assert_almost_equal(exp_values, reference_exp_values, decimal=5)

    def test_get_exp_value_from_statevector_h2(self):
        """ Get expectation value of large circuits and qubit Hamiltonians corresponding to molecules.
            Molecule: H2 sto-3g = [("H", (0., 0., 0.)), ("H", (0., 0., 0.741377))]
        """
        qubit_operator = load_operator("mol_H2_qubitham.data", data_directory=path_data, plain_text=True)

        with open(f"{path_data}/H2_UCCSD.qasm", "r") as circ_handle:
            openqasm_circ = circ_handle.read()

        abs_circ = translator._translate_openqasm2abs(openqasm_circ)
        expected = -1.1372704
        test_fail = False

        for b in installed_sv_simulator:
            sim = Simulator(target=b)
            tstart = time.time()
            energy = sim.get_expectation_value(qubit_operator, abs_circ)
            tstop = time.time()
            print(f"H2 get exp value with {b:10s} returned {energy:.7f} \t Elapsed: {tstop - tstart:.3f} s.")

            try:
                self.assertAlmostEqual(energy, expected, delta=1e-5)
            except AssertionError:
                test_fail = True
                print(f"{self._testMethodName} : Assertion failed {b} (result = {energy:.7f}, expected = {expected})")
        if test_fail:
            assert False

    def test_get_exp_value_from_initial_statevector_h2(self):
        """ Get expectation value of large circuits and qubit Hamiltonians corresponding to molecules.
            Molecule: H2 sto-3g = [("H", (0., 0., 0.)), ("H", (0., 0., 0.741377))]
            Generate statevector first and then get_expectation value from statevector and empty circuit.
        """
        qubit_operator = load_operator("mol_H2_qubitham.data", data_directory=path_data, plain_text=True)

        with open(f"{path_data}/H2_UCCSD.qasm", "r") as circ_handle:
            openqasm_circ = circ_handle.read()

        abs_circ = translator._translate_openqasm2abs(openqasm_circ)
        expected = -1.1372704
        test_fail = False

        for b in installed_sv_simulator:
            sim = Simulator(target=b)
            tstart = time.time()
            _, statevector = sim.simulate(abs_circ, return_statevector=True)
            energy = sim.get_expectation_value(qubit_operator, Circuit(n_qubits=abs_circ.width),
                                               initial_statevector=statevector)
            tstop = time.time()
            print(f"H2 get exp value with {b:10s} returned {energy:.7f} \t Elapsed: {tstop - tstart:.3f} s.")

            try:
                self.assertAlmostEqual(energy, expected, delta=1e-5)
            except AssertionError:
                test_fail = True
                print(f"{self._testMethodName} : Assertion failed {b} (result = {energy:.7f}, expected = {expected})")
        if test_fail:
            assert False

    def test_get_exp_value_from_statevector_h4(self):
        """ Get expectation value of large circuits and qubit Hamiltonians corresponding to molecules.
            Molecule: H4 sto-3g
            H4 = [['H', [0.7071067811865476,   0.0,                 0.0]],
                  ['H', [0.0,                  0.7071067811865476,  0.0]],
                  ['H', [-1.0071067811865476,  0.0,                 0.0]],
                  ['H', [0.0,                 -1.0071067811865476,  0.0]]]
        """
        qubit_operator = load_operator("mol_H4_qubitham.data", data_directory=path_data, plain_text=True)

        with open(f"{path_data}/H4_UCCSD.qasm", "r") as circ_handle:
            openqasm_circ = circ_handle.read()

        abs_circ = translator._translate_openqasm2abs(openqasm_circ)
        expected = -1.9778374
        test_fail = False

        for b in installed_sv_simulator:
            sim = Simulator(target=b)
            tstart = time.time()
            energy = sim.get_expectation_value(qubit_operator, abs_circ)
            tstop = time.time()
            print(f"H4 get exp value with {b:10s} returned {energy:.7f} \t Elapsed: {tstop - tstart:.3f} s.")

            try:
                self.assertAlmostEqual(energy, expected, delta=1e-5)
            except AssertionError:
                test_fail = True
                print(f"{self._testMethodName} : Assertion failed {b} (result = {energy:.7f}, expected = {expected})")
        if test_fail:
            assert False

    @unittest.skipIf("qulacs" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_get_exp_value_from_statevector_with_shots_h2(self):
        """ Get expectation value of large circuits and qubit Hamiltonians corresponding to molecules.
            Molecule: H2 sto-3g = [("H", (0., 0., 0.)), ("H", (0., 0., 0.741377))]
            The result is computed using samples ("shots") drawn form a statevector simulator here. This is the kind
            of results we could expect from a noiseless QPU.
        """
        qubit_operator = load_operator("mol_H2_qubitham.data", data_directory=path_data, plain_text=True)

        with open(f"{path_data}/H2_UCCSD.qasm", "r") as circ_handle:
            openqasm_circ = circ_handle.read()
        abs_circ = translator._translate_openqasm2abs(openqasm_circ)

        simulator = Simulator(target="qulacs", n_shots=10 ** 6)
        expected = -1.1372704

        energy = simulator.get_expectation_value(qubit_operator, abs_circ)
        self.assertAlmostEqual(energy, expected, delta=1e-3)

    def test_get_exp_value_empty_circuit(self):
        """ If the circuit is empty and we have a non-zero number of qubits, frequencies just only show all-|0> state
        observed and compute the expectation value using these frequencies """

        empty_circuit = Circuit([], n_qubits=2)
        identity_circuit = Circuit([Gate('X', 0), Gate('X', 1)] * 2)

        for b in installed_sv_simulator:
            simulator = Simulator(target=b)
            for op in [op1, op2]:
                exp_value_empty = simulator.get_expectation_value(op, empty_circuit)
                exp_value_identity = simulator.get_expectation_value(op, identity_circuit)
                np.testing.assert_almost_equal(exp_value_empty, exp_value_identity, decimal=8)

    def test_get_exp_value_complex(self):
        """ Get expectation value of qubit operator with complex coefficients """

        for b in installed_sv_simulator:
            simulator = Simulator(target=b)

            # Return complex expectation value corresponding to linear combinations of real and imaginary parts
            op_c = op1 + 1.0j * op2
            exp_c = simulator.get_expectation_value(op_c, circuit3)
            exp_r1 = simulator.get_expectation_value(op1, circuit3)
            exp_r2 = simulator.get_expectation_value(op2, circuit3)
            self.assertAlmostEqual(exp_c.real, exp_r1, delta=1.e-12)
            self.assertAlmostEqual(exp_c.imag, exp_r2, delta=1.e-12)

            # Edge case: all coefficients are complex but with imaginary part null: exp value must return a float
            op_c = op1 + 0.j * op1
            exp_c = simulator.get_expectation_value(op_c, circuit3)
            assert (type(exp_c) in {float, np.float64} and exp_c == exp_r1)

    def test_get_exp_value_from_frequencies(self):
        """ Test the method computing the expectation value from frequencies, with a given simulator """

        for b in installed_sv_simulator:
            simulator = Simulator(target=b)
            exp_values = np.zeros((len(circuits), len(ops)), dtype=float)
            for i, circuit in enumerate(circuits):
                for j, op in enumerate(ops):
                    exp_values[i][j] = simulator._get_expectation_value_from_frequencies(op, circuit)
            np.testing.assert_almost_equal(exp_values, reference_exp_values, decimal=5)


class TestSimulateMisc(unittest.TestCase):

    def test_n_shots_needed(self):
        """
            Raise an error if user chooses a target backend that does not provide access to a statevector and
            also does not provide a number of shots for the simulation.
        """
        self.assertRaises(ValueError, Simulator, target="qdk")

    @unittest.skipIf("qdk" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_simulate_qdk(self):
        """
            Must return correct frequencies for simulation of different quantum circuits.
            The accuracy is correlated to the number of shots taken in the simulation.
            Backend: qdk.
        """
        simulator = Simulator(target="qdk", n_shots=10**4)
        for i, circuit in enumerate(circuits):
            frequencies, _ = simulator.simulate(circuit)
            assert_freq_dict_almost_equal(ref_freqs[i], frequencies, atol=1e-1)

    @unittest.skipIf("qdk" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_get_exp_value_from_frequencies_qdk(self):
        """ Test specific to QDK to ensure results are not impacted by code specific to frequency computation
            as well as the recompilation of the Q# file used in successive simulations """

        simulator = Simulator(target="qdk", n_shots=10**4)
        exp_values = np.zeros((len(ops)), dtype=float)
        for j, op in enumerate(ops):
            exp_values[j] = simulator.get_expectation_value(op, circuit3)
        np.testing.assert_almost_equal(exp_values, reference_exp_values[2], decimal=1)

    def test_get_exp_value_from_frequencies_oneterm(self):
        """ Test static method computing the expectation value of one term, when the results of a simulation
         are being provided as input. """

        term, coef = ((0, 'Z'),), 1.0  # Data as presented in Openfermion's QubitOperator.terms attribute
        exp_value = coef * Simulator.get_expectation_value_from_frequencies_oneterm(term, ref_freqs[2])
        np.testing.assert_almost_equal(exp_value, -0.41614684, decimal=5)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python wrappers facilitating quantum experiment submission, monitoring and
post-processing, through QEMIST Cloud.

Users are expected to set the environment variables QEMIST_AUTH_TOKEN and
QEMIST_PROJECT_ID with values retrieved from their QEMIST Cloud dashboard.
"""

try:
    import qemist_client.util as qclient_util
except ModuleNotFoundError:
    print("qemist_client python package not found (optional dependency for hardware experiment submission)")


def job_submit(circuit, n_shots, backend):
    """Job submission to run a circuit on quantum hardware.

    Args:
        circuit: a quantum circuit in the abstract format.
        n_shots (int): the number of shots.
        backend (str): the identifier string for the desired backend.

    Returns:
        int: A problem handle / job ID that can be used to retrieve the result
            or cancel the problem.
    """

    # Serialize circuit data
    circuit_data = circuit.serialize()

    # Build option dictionary
    job_options = {'shots': n_shots, 'backend': backend}

    # Submit the problem
    qemist_cloud_job_id = qclient_util.solve_quantum_circuits_async(serialized_fragment=circuit_data,
                                                                    serialized_solver=job_options)[0]
    return qemist_cloud_job_id


def job_status(qemist_cloud_job_id):
    """Returns the current status of the problem, as a string. Possible values:
    ready, in_progress, complete, cancelled.

    Args:
        qemist_cloud_job_id (int): problem handle / job identifier.

    Returns:
        str: current status of the problem, as a string.
    """
    res = qclient_util.get_problem_statuses(qemist_cloud_job_id)

    return res


def job_cancel(qemist_cloud_job_id):
    """Cancels the job matching the input job id, if done in time before it
    starts.

    Args:
        qemist_cloud_job_id (int): problem handle / job identifier.

    Returns:
        dict: cancelled problems / subproblems.
    """
    res = qclient_util.cancel_problems(qemist_cloud_job_id)
    # TODO: If res is coming out as an error code, we should raise an error

    return res


def job_result(qemist_cloud_job_id):
    """Blocks until the job results are available. Returns a tuple containing
    the histogram of frequencies, and also the more in-depth raw data from the
    cloud services provider as a nested dictionary

    Args:
        qemist_cloud_job_id (int): problem handle / job identifier.

    Returns:
        dict: Histogram of measurement frequencies.
        dict: The cloud provider raw data.
    """

    try:
        qclient_util.monitor_problem_status(problem_handles=[qemist_cloud_job_id], verbose=False)

    except KeyboardInterrupt:
        print(f"\nYour problem is still running with id {qemist_cloud_job_id}.\n")
        command = input("Type 'cancel' and return to cancel your problem."
                        "Type anything else to disconnect but keep the problem running.\n")
        if command.lower() == "cancel":
            ret = job_cancel(qemist_cloud_job_id)
            print("Problem cancelled.", ret)
        else:
            print(f"Reconnect and block until the problem is complete with "
                  f"qemist_client.util.monitor_problem_status({qemist_cloud_job_id}).\n\n")
        raise

    except Exception:
        print(f"\n\nYour problem is still running with handle {qemist_cloud_job_id}.\n"
              f"Cancel the problem with qemist_client.util.cancel_problems({qemist_cloud_job_id}).\n"
              f"Reconnect and block until the problem is complete with qemist_client.util.monitor_problem_status({qemist_cloud_job_id}).\n\n")
        raise

    # Once a result is available, retrieve it.
    # If the util module is not found earlier, an error has been raised.
    output = qclient_util.get_results(qemist_cloud_job_id)

    # Amazon Braket: parsing of output
    freqs = output['results']['measurement_probabilities']
    raw_data = output

    return freqs, raw_data


def job_estimate(circuit, n_shots, backend=None):
    """Returns an estimate of the cost of running an experiment, for a specified backend
    or all backends available. Some service providers care about the
    complexity / structure of the input quantum circuit, some do not.

    The backend identifier strings that a user can provide as argument can be obtained
    by calling this function without specifying a backend. They appear as keys in
    the returned dictionary. These strings may change with time, as we adjust to the
    growing cloud quantum offer (services and devices).

    Args:
        circuit (Circuit): the abstract circuit to be run on the target device.
        n_shots (int): number of shots in the expriment.
        backend (str): the identifier string for the desired backend.

    Returns:
        dict: Returns dict of prices in USD. If backend is not None, dictionary
        contains the cost for running the desired job. If backend is None,
        returns dictionary of prices for all supported backends.
    """

    # Serialize circuit data
    circuit_data = circuit.serialize()

    # Build option dictionary
    job_options = {'shots': n_shots}
    if backend:
        job_options['backend'] = backend

    price_estimate = qclient_util.check_qpu_cost(circuit_data, job_options)

    return price_estimate



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import numpy as np

from tangelo.toolboxes.operators import QubitOperator, MultiformOperator
from tangelo.toolboxes.operators.multiformoperator import do_commute

qu_op_xyz = QubitOperator("X0 Y1 Z2", 1.)
qu_op_zyz = QubitOperator("Z0 Y1 Z2", 1.)

int_op_xyz = np.array([[2, 3, 1]])
bin_op_xyz = np.array([[1, 1, 0, 0, 1, 1]])


class MultiformOperatorUtilitiesTest(unittest.TestCase):

    def test_do_commute(self):
        """Test is_commuting function."""

        commute_xyz_xyz = do_commute(MultiformOperator.from_qubitop(qu_op_xyz),
                                     MultiformOperator.from_qubitop(qu_op_xyz))
        self.assertTrue(commute_xyz_xyz)

        commute_xyz_zyz = do_commute(MultiformOperator.from_qubitop(qu_op_xyz),
                                     MultiformOperator.from_qubitop(qu_op_zyz))
        self.assertFalse(commute_xyz_zyz)


class MultiformOperatorTest(unittest.TestCase):

    def test_instantiate_from_QubitOperator(self):
        """Test initialization of MultiformOperator class with a QubitOperator."""

        hydrib_op = MultiformOperator.from_qubitop(qu_op_xyz)

        self.assertDictEqual(hydrib_op.terms, qu_op_xyz.terms)
        np.testing.assert_allclose(hydrib_op.integer, int_op_xyz)
        np.testing.assert_allclose(hydrib_op.binary, bin_op_xyz)

    def test_instantiate_from_integers(self):
        """Test initialization of MultiformOperator class with integers."""

        hydrib_op = MultiformOperator.from_integerop(int_op_xyz, [1.])

        self.assertDictEqual(hydrib_op.terms, qu_op_xyz.terms)
        np.testing.assert_allclose(hydrib_op.integer, int_op_xyz)
        np.testing.assert_allclose(hydrib_op.binary, bin_op_xyz)

    def test_instantiate_from_binary(self):
        """Test initialization of MultiformOperator class with binary numbers."""

        hydrib_op = MultiformOperator.from_binaryop(bin_op_xyz, [1.])

        self.assertDictEqual(hydrib_op.terms, qu_op_xyz.terms)
        np.testing.assert_allclose(hydrib_op.integer, int_op_xyz)
        np.testing.assert_allclose(hydrib_op.binary, bin_op_xyz)

    def test_multiply(self):
        """Test multiplication of 2 MultiformOperators."""

        hydrib_op_a = MultiformOperator.from_qubitop(qu_op_xyz)
        hydrib_op_b = MultiformOperator.from_qubitop(qu_op_zyz)

        aa = hydrib_op_a * hydrib_op_a
        np.testing.assert_allclose(aa.integer, np.array([[0, 0, 0]]))
        np.testing.assert_allclose(aa.factors, np.array([1.]))

        ab = hydrib_op_a * hydrib_op_b
        np.testing.assert_allclose(ab.integer, np.array([[3, 0, 0]]))
        np.testing.assert_allclose(ab.factors, np.array([-1.j]))

        ba = hydrib_op_b * hydrib_op_a
        np.testing.assert_allclose(ba.integer, np.array([[3, 0, 0]]))
        np.testing.assert_allclose(ba.factors, np.array([1.j]))

    def test_collapse(self):
        """Test collapse function with a given operator with duplicate terms."""

        int_op = np.array([
            [0, 1, 2, 3],
            [0, 1, 2, 3]
        ])

        collapsed_int_op, collapsed_factors = MultiformOperator.collapse(int_op, factors=np.ones(int_op.shape[0]))

        np.testing.assert_array_equal([[0, 1, 2, 3]], collapsed_int_op)
        np.testing.assert_array_equal([2.], collapsed_factors)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""This module defines the fermionic operators that can be used to obtain
expectation values of commonly used quantum numbers. The available operators are
1) N: number of electrons
2) Sz: The spin z-projection Sz|\psi>=m_s|\psi>
3) S^2: The spin quantum number S^2|\psi>=s(s+1)|\psi> associated with spin
angular momentum which allows one to decide whether the state has the correct
properties.
"""

from tangelo.toolboxes.ansatz_generator._general_unitary_cc import get_spin_ordered
from tangelo.toolboxes.operators import normal_ordered, list_to_fermionoperator


def number_operator(n_orbs, up_then_down=False):
    r"""Function to generate the normal ordered number operator as a
    FermionOperator.

    Args:
        n_orbs (int): number of orbitals in the fermion basis (this is number of
            spin-orbitals divided by 2).
        up_then_down: The ordering of the spin orbitals. qiskit (True)
            openfermion (False).

    Returns:
        FermionOperator: The number operator penalty \hat{N}.
    """

    all_terms = number_operator_list(n_orbs, up_then_down)
    num_op = list_to_fermionoperator(all_terms)

    return normal_ordered(num_op)


def number_operator_list(n_orbs, up_then_down=False):
    r"""Function to generate the normal ordered number operator as a list.

    Args:
        n_orbs (int): number of orbitals in the fermion basis (this is number of
            spin-orbitals divided by 2).
        up_then_down (bool): The ordering of the spin orbitals. qiskit (True)
            openfermion (False).

    Returns:
        list: The number operator penalty \hat{N}.
    """

    all_terms = list()
    for i in range(n_orbs):
        up, dn = get_spin_ordered(n_orbs, i, i, up_down=up_then_down)  # get spin-orbital indices
        all_terms.extend([[((up[0], 1), (up[1], 0)), 1], [((dn[0], 1), (dn[1], 0)), 1]])
    return all_terms


def spinz_operator(n_orbs, up_then_down=False):
    r"""Function to generate the normal ordered Sz operator.

    Args:
        n_orbs (int): number of orbitals in the fermion basis (this is number of
            spin-orbitals divided by 2).
        up_then_down (bool): The ordering of the spin orbitals. qiskit (True)
            openfermion (False).

    Returns:
        FermionOperator: The Sz operator \hat{Sz}.
    """

    all_terms = spinz_operator_list(n_orbs, up_then_down)
    spinz_op = list_to_fermionoperator(all_terms)

    return normal_ordered(spinz_op)


def spinz_operator_list(n_orbs, up_then_down=False):
    r"""Function to generate the normal ordered Sz operator as a list.

    Args:
        n_orbs (int): number of orbitals in the fermion basis (this is number of
            spin-orbitals divided by 2).
        up_then_down (bool): The ordering of the spin orbitals. qiskit (True)
            openfermion (False).

    Returns:
        list: The Sz operator \hat{Sz}.
    """

    all_terms = list()
    for i in range(n_orbs):
        up, dn = get_spin_ordered(n_orbs, i, i, up_down=up_then_down)  # get spin-orbital indices
        all_terms.extend([[((up[0], 1), (up[1], 0)), 1/2], [((dn[0], 1), (dn[1], 0)), -1/2]])
    return all_terms


def spin2_operator(n_orbs, up_then_down=False):
    r"""Function to generate the normal ordered S^2 operator as a
    FermionOperator.

    Args:
        n_orbs (int): number of orbitals in the fermion basis (this is number of
            spin-orbitals divided by 2).
        up_then_down (bool): The ordering of the spin orbitals. qiskit (True)
            openfermion (False).

    Returns:
        FermionOperator: The S^2 operator \hat{S}^2.
    """

    all_terms = spin2_operator_list(n_orbs, up_then_down)
    spin2_op = list_to_fermionoperator(all_terms)

    return normal_ordered(spin2_op)


def spin2_operator_list(n_orbs, up_then_down=False):
    r"""Function to generate the normal ordered S^2 operator as a list.

    Args:
        n_orbs (int): number of orbitals in the fermion basis (this is number of
            spin-orbitals divided by 2).
        up_then_down (bool): The ordering of the spin orbitals. qiskit (True)
            openfermion (False).

    Returns:
        list: The S^2 operator \hat{S}^2.
    """

    all_terms = list()
    for i in range(n_orbs):
        up, dn = get_spin_ordered(n_orbs, i, i, up_down=up_then_down)  # get spin-orbital indices
        all_terms.extend([[((up[0], 1), (up[1], 0), (up[0], 1), (up[1], 0)), 1/4],
                         [((dn[0], 1), (dn[1], 0), (dn[0], 1), (dn[1], 0)), 1/4],
                         [((up[0], 1), (up[1], 0), (dn[0], 1), (dn[1], 0)), -1/4],
                         [((dn[0], 1), (dn[1], 0), (up[0], 1), (up[1], 0)), -1/4],
                         [((up[0], 1), (dn[1], 0), (dn[0], 1), (up[1], 0)), 1/2],
                         [((dn[0], 1), (up[1], 0), (up[0], 1), (dn[1], 0)), 1/2]])
        for j in range(n_orbs):
            if (i != j):
                up2, dn2 = get_spin_ordered(n_orbs, j, j, up_down=up_then_down)
                all_terms.extend([[((up[0], 1), (up[1], 0), (up2[0], 1), (up2[1], 0)), 1/4],
                                 [((dn[0], 1), (dn[1], 0), (dn2[0], 1), (dn2[1], 0)), 1/4],
                                 [((up[0], 1), (up[1], 0), (dn2[0], 1), (dn2[1], 0)), -1/4],
                                 [((dn[0], 1), (dn[1], 0), (up2[0], 1), (up2[1], 0)), -1/4],
                                 [((up[0], 1), (dn[1], 0), (dn2[0], 1), (up2[1], 0)), 1/2],
                                 [((dn[0], 1), (up[1], 0), (up2[0], 1), (dn2[1], 0)), 1/2]])
    return all_terms



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for statevector mapping methods, which carry a numpy array indicating
fermionic occupation of reference state into qubit representation.
"""


import unittest
from itertools import combinations

import numpy as np

from tangelo.toolboxes.qubit_mappings.statevector_mapping import get_vector, vector_to_circuit
from tangelo.toolboxes.molecular_computation.molecule import SecondQuantizedMolecule
from tangelo.molecule_library import mol_H4_sto3g, mol_H4_cation_sto3g, xyz_H2O
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping
from tangelo.linq import Simulator

sim = Simulator()
spin_ind_transforms = ["BK", "JW", "JKMN"]
spin_dep_transforms = ["SCBK"]
all_transforms = spin_ind_transforms + spin_dep_transforms


class TestVector(unittest.TestCase):

    def test_jw_value(self):
        """Check that Jordan-Wigner mapping returns correct vector, for both
        default spin orderings.
        """
        vector = np.array([1, 1, 1, 1, 0, 0, 0, 0])
        vector_updown = np.array([1, 1, 0, 0, 1, 1, 0, 0])

        output_jw = get_vector(vector.size, sum(vector), mapping="jw", up_then_down=False)
        output_jw_updown = get_vector(vector.size, sum(vector), mapping="jw", up_then_down=True)
        self.assertEqual(np.linalg.norm(vector - output_jw), 0.0)
        self.assertEqual(np.linalg.norm(vector_updown - output_jw_updown), 0.0)

    def test_bk_value(self):
        """Check that Bravyi-Kitaev mapping returns correct vector, for both
        default spin orderings.
        """
        vector = np.array([1, 1, 1, 1, 0, 0, 0, 0])
        vector_bk = np.array([1, 0, 1, 0, 0, 0, 0, 0])
        vector_bk_updown = np.array([1, 0, 0, 0, 1, 0, 0, 0])

        output_bk = get_vector(vector.size, sum(vector), mapping="bk", up_then_down=False)
        output_bk_updown = get_vector(vector.size, sum(vector), mapping="bk", up_then_down=True)
        self.assertEqual(np.linalg.norm(vector_bk - output_bk), 0.0)
        self.assertEqual(np.linalg.norm(vector_bk_updown - output_bk_updown), 0.0)

    def test_scbk_value(self):
        """Check that symmetry-conserving Bravyi-Kitaev mapping returns correct
        vector.
        """
        vector = np.array([1, 0, 0, 1, 0, 0])

        output_bk = get_vector(8, 4, mapping="SCBK", up_then_down=True)
        self.assertEqual(np.linalg.norm(vector - output_bk), 0.0)

    def test_circuit_build(self):
        """Check circuit width and size (number of X gates)."""
        vector = np.array([1, 1, 1, 1, 0, 0, 1, 1])
        circuit = vector_to_circuit(vector)
        self.assertEqual(circuit.size, sum(vector))
        self.assertEqual(circuit.width, vector.size)

    def test_all_same_energy_mol_H4_sto3g_and_mol_H2O_sto3g(self):
        """Check that all mappings return statevectors that have the same energy expectation
        for an even number of electrons and various spins. Molecules tested are H4, and H2O
        with frozen_orbitals=[0, 7] which failed previously for scbk"""
        mols = [mol_H4_sto3g, SecondQuantizedMolecule(xyz_H2O, 0, 0, "sto-3g", frozen_orbitals=[0, 7])]
        circuits = dict()
        qu_ops = dict()
        for mol in mols:
            ferm_op = mol.fermionic_hamiltonian
            qu_ops = dict()
            for transform in spin_ind_transforms:
                qu_ops[transform] = fermion_to_qubit_mapping(ferm_op,
                                                             transform,
                                                             mol.n_active_sos,
                                                             mol.n_active_electrons,
                                                             up_then_down=True)

            # Test for spin 0, 2, and 4
            for spin in range(3):
                # Get circuits for transforms not dependant on spin
                for transform in all_transforms:
                    vector = get_vector(mol.n_active_sos,
                                        mol.n_active_electrons,
                                        mapping=transform,
                                        up_then_down=True,
                                        spin=spin*2)
                    circuits[transform] = vector_to_circuit(vector)

                for transform in spin_dep_transforms:
                    qu_ops[transform] = fermion_to_qubit_mapping(ferm_op,
                                                                 'SCBK',
                                                                 mol.n_active_sos,
                                                                 mol.n_active_electrons,
                                                                 up_then_down=True,
                                                                 spin=spin*2)

                energies = dict()
                for transform in all_transforms:
                    energies[transform] = sim.get_expectation_value(qu_ops[transform], circuits[transform])

                for t1, t2 in combinations(all_transforms, 2):
                    self.assertAlmostEqual(energies[t1], energies[t2], places=7, msg=f"Failed for {t1} vs {t2} for spin={spin*2}")

    def test_all_same_energy_mol_H4_cation_sto3g(self):
        """Check that all mappings return statevectors that have the same energy expectation
        for an odd number of electrons and various spins"""
        ferm_op = mol_H4_cation_sto3g.fermionic_hamiltonian
        circuits = dict()
        qu_ops = dict()
        for transform in spin_ind_transforms:
            qu_ops[transform] = fermion_to_qubit_mapping(ferm_op,
                                                         transform,
                                                         mol_H4_cation_sto3g.n_active_sos,
                                                         mol_H4_cation_sto3g.n_active_electrons,
                                                         up_then_down=True)

        # Test for spin 1 and 3
        for spin in range(2):
            for transform in all_transforms:
                vector = get_vector(mol_H4_cation_sto3g.n_active_sos,
                                    mol_H4_cation_sto3g.n_active_electrons,
                                    mapping=transform,
                                    up_then_down=True,
                                    spin=spin*2+1)
                circuits[transform] = vector_to_circuit(vector)
            for transform in spin_dep_transforms:
                qu_ops[transform] = fermion_to_qubit_mapping(ferm_op,
                                                             transform,
                                                             mol_H4_cation_sto3g.n_active_sos,
                                                             mol_H4_cation_sto3g.n_active_electrons,
                                                             up_then_down=True,
                                                             spin=spin*2+1)

            energies = dict()
            for transform in all_transforms:
                energies[transform] = sim.get_expectation_value(qu_ops[transform], circuits[transform])

            for t1, t2 in combinations(all_transforms, 2):
                self.assertAlmostEqual(energies[t1], energies[t2], places=7, msg=f"Failed for {t1} vs {t2} for spin={spin*2}")


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module implements functions to create the direct interaction set (DIS)
of generators for the qubit coupled cluster (QCC) ansatz and is based on Ref. 1.
The DIS consists of generator groups that are characterized by the magnitude of
the gradient of the QCC energy functional with respect to a variational parameter
tau, |dEQCC/dtau|. Nonzero values of |dEQCC/dtau| imply that an individual
generator will contribute to variational energy lowering. The number of DIS groups
cannot exceed the number of Hamiltonian terms, N, and each DIS group contains
2^nq - 1 generators, where nq is the number of qubits, all with identical values
of |dEQCC/dtau|. By constructing the DIS, it is possible to identify O(N * (2^nq - 1))
generators that are strong energy-lowering candidates for the QCC ansatz at a
cost of O(N) gradient evaluations. In contrast, a brute-force strategy requires
O(4^nq) gradient evaluations.

Refs:
    1. I. G. Ryabinkin, R. A. Lang, S. N. Genin, and A. F. Izmaylov.
        J. Chem. Theory Comput. 2020, 16, 2, 1055–1063.
"""

from math import sin, cos
from itertools import combinations

from openfermion import commutator

from tangelo.toolboxes.operators.operators import QubitOperator
from tangelo.toolboxes.ansatz_generator._qubit_mf import get_op_expval


def construct_dis(qubit_ham, pure_var_params, deqcc_dtau_thresh):
    """ Construct the direct interaction set (DIS) of QCC generators as follows:
    1. Identify the flip indices of all Hamiltonian terms and group terms by flip indices.
    2. Construct a representative generator using flip indices from each candidate DIS group
       and evaluate dEQCC/dtau for all Hamiltonian terms.
    3. For each candidate DIS group, sum dEQCC/dtau for all Hamiltonian terms.
       If |dEQCC/dtau| >= thresh add the candidate DIS group to the DIS.
    4. For all DIS groups, create a complete set of generators made from Pauli X and and an
       odd number of Y operators.

    Args:
        qubit_ham (QubitOperator): A qubit Hamiltonian.
        pure_var_params (numpy array of float): A purified QMF variational parameter set.
        deqcc_dtau_thresh (float): Threshold for |dEQCC/dtau| so that a candidate group is added
            to the DIS if |dEQCC/dtau| >= deqcc_dtau_thresh for a generator.

    Returns:
        list of list: the DIS of QCC generators.
    """

    # Use a qubit Hamiltonian and purified QMF parameter set to construct the DIS
    dis, dis_groups = [], get_dis_groups(qubit_ham, pure_var_params, deqcc_dtau_thresh)
    if dis_groups:
        for dis_group in dis_groups:
            dis_group_idxs = [int(idxs) for idxs in dis_group[0].split(" ")]
            dis_group_gens = get_gens_from_idxs(dis_group_idxs)
            # for now just grab the first generator; eventually add capability to
            # allow the user to select which generators to use.
            dis.append(dis_group_gens[0])
    else:
        raise ValueError(f"The DIS is empty: there are no candidate DIS groups where "
                         f"|dEQCC/dtau| >= {deqcc_dtau_thresh} a.u. Terminate simulation.\n")
    return dis


def get_dis_groups(qubit_ham, pure_var_params, deqcc_dtau_thresh):
    """Construct unique DIS groups characterized by the flip indices and |dEQCC/dtau|.

    Args:
        qubit_ham (QubitOperator): A qubit Hamiltonian.
        pure_var_params (numpy array of float): A purified QMF variational parameter set.
        deqcc_dtau_thresh (float): Threshold for |dEQCC/dtau| so that a candidate group is added
            to the DIS if |dEQCC/dtau| >= deqcc_dtau_thresh for a generator.

    Returns:
        list of tuple: the DIS group flip indices (str) and signed value of dEQCC/dtau (float).
    """

    # Get the flip indices from qubit_ham and compute the gradient dEQCC/dtau
    qham_gen = ((qham_items[0], (qham_items[1], pure_var_params))
                 for qham_items in qubit_ham.terms.items())
    flip_idxs = list(filter(None, (get_idxs_deriv(q_gen[0], *q_gen[1]) for q_gen in qham_gen)))

    # Group Hamiltonian terms with the same flip indices and sum of the signed dEQCC/tau values
    candidates = dict()
    for idxs in flip_idxs:
        deriv_old = candidates.get(idxs[0], 0.)
        candidates[idxs[0]] = idxs[1] + deriv_old

    # Return a sorted list of flip indices and signed dEQCC/dtau values for each DIS group
    dis_groups = [idxs_deriv for idxs_deriv in candidates.items()
                  if abs(idxs_deriv[1]) >= deqcc_dtau_thresh]
    return sorted(dis_groups, key=lambda deriv: abs(deriv[1]), reverse=True)


def get_idxs_deriv(qham_term, *qham_qmf_data):
    """Find the flip indices of a qubit Hamiltonian term by identifying the indices
    of any X and Y operators that are present. A representative generator is then
    built with a Pauli Y operator acting on the first flip index and then Pauli X
    operators acting on the remaining flip indices. Then dEQCC/dtau is evaluated as
    dEQCC_dtau = -i/2 <QMF|[H, gen]|QMF> = -i <QMF|H * gen|QMF>.

    Args:
        qham_term (tuple of tuple): The Pauli operators and indices of a QubitOperator term.
        qham_qmf_data (tuple): The coefficient of a QubitOperator term and a purified QMF
            variational parameter set (numpy array of float).

    Returns:
        tuple or None: return a tuple of the flip indices (str) and the signed value of
            dEQCC/dtau (float) if at least two flip indices were found. Otherwise return None.
    """

    coef, pure_params = qham_qmf_data
    idxs, gen_tup, idxs_deriv = "", tuple(), None
    for pauli_factor in qham_term:
        # The indices of X and Y operators are flip indices
        idx, pauli_op = pauli_factor
        if "X" in pauli_op or "Y" in pauli_op:
            gen = (idx, "Y") if idxs == "" else (idx, "X")
            idxs = idxs + f" {idx}" if idxs != "" else f"{idx}"
            gen_tup += (gen, )
    # Generators must have at least two flip indices
    if len(gen_tup) > 1:
        qham_gen_comm = QubitOperator(qham_term, -1j * coef)
        qham_gen_comm *= QubitOperator(gen_tup, 1.)
        deriv = get_op_expval(qham_gen_comm, pure_params).real
        idxs_deriv = (idxs, deriv)
    return idxs_deriv


def get_gens_from_idxs(group_idxs):
    """Given the flip indices of a DIS group, create all possible Pauli words made
    from Pauli X and an odd number of Y operators acting on qubits indexed by the
    flip indices.

    Args:
        group_idxs (str): A set of flip indices for a DIS group.

    Returns:
        list of QubitOperator: DIS group generators.
    """

    dis_group_gens = []
    for n_y in range(1, len(group_idxs), 2):
        # Create combinations of odd numbers of flip indices for the Pauli Y operators
        for xy_idx in combinations(group_idxs, n_y):
            # If a flip index idx matches xy_idx, add a Y operator
            gen_list = [(idx, "Y") if idx in xy_idx else (idx, "X") for idx in group_idxs]
            dis_group_gens.append(QubitOperator(tuple(gen_list), 1.))
    return dis_group_gens


def build_qcc_qubit_op(dis_gens, amplitudes):
    """Returns the QCC operator by selecting n_var_params generators from the DIS.
    The QCC operator is constructed as a linear combination of generators using the
    parameter set {tau} as coefficients: QCC operator = -0.5 * SUM_k P_k * tau_k.
    The exponentiated QCC operator, U = PROD_k exp(-0.5j * tau_k * P_k), is used to
    build the circuit.

    Args:
        dis_gens (list of QubitOperator): The list of QCC Pauli word generators
            selected from a user-specified number of characteristic DIS groups.
        amplitudes (list or numpy array of float): The QCC variational parameters
            arranged such that their ordering matches the order of dis_gens.

    Returns:
        QubitOperator: QCC ansatz operator.
    """

    qubit_op = QubitOperator.zero()
    for i, dis_gen in enumerate(dis_gens):
        qubit_op -= 0.5 * amplitudes[i] * dis_gen
    qubit_op.compress()
    return qubit_op


def qcc_op_dress(qubit_op, dis_gens, amplitudes):
    """Performs canonical transformation of a qubit operator with the set of QCC
    generators and amplitudes for the current iteration. For an operator with M terms
    each transformation results in exponential growth of the number terms. This growth
    can be approximated as M * (3 / 2) ^ n_g, where n_g is the number of QCC generators
    selected for the ansatz at the current iteration.

    Args:
        qubit_op (QubitOperator): A qubit operator (e.g., a molecular Hamiltonian or the
            electronic spin and number operators) that was previously dressed by canonical
            transformation with the QCC generators and amplitudes at the current iteration.
        dis_gens (list of QubitOperator): The list of QCC Pauli word generators
            selected from a user-specified number of characteristic DIS groups.
        amplitudes (list or numpy array of float): The QCC variational parameters
            arranged such that their ordering matches the ordering of dis_gens.

    Returns:
        QubitOperator: Dressed qubit operator.
    """

    for i, gen in enumerate(dis_gens):
        comm = commutator(qubit_op, gen)
        qubit_op += .5 * ((1. - cos(amplitudes[i])) * gen - 1j * sin(amplitudes[i])) * comm
    qubit_op.compress()
    return qubit_op



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from tangelo.molecule_library import mol_H2O_321g
from tangelo.toolboxes.molecular_computation.frozen_orbitals import get_frozen_core, get_orbitals_excluding_homo_lumo


class FrozenOrbitalsTest(unittest.TestCase):

    def test_get_frozen_core(self):
        """Verify if the frozen orbital suggestions are consistent with chemical
        intuition.
        """

        frozen_h2o = get_frozen_core(mol_H2O_321g)
        self.assertEqual(frozen_h2o, 1)

    def test_get_homo_lumo(self):
        """Verify if the HOMO-LUMO suggestions are consistent with the provided
        parameters.
        """

        # Getting HOMO-LUMO.
        frozen_homo_lumo = get_orbitals_excluding_homo_lumo(mol_H2O_321g)
        self.assertEqual(frozen_homo_lumo, [0, 1, 2, 3, 6, 7, 8, 9, 10, 11, 12])

        # Active space from HOMO-2 to LUMO+4
        frozen_homo2_lumo4 = get_orbitals_excluding_homo_lumo(mol_H2O_321g, 2, 4)
        self.assertEqual(frozen_homo2_lumo4, [0, 1, 10, 11, 12])


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import itertools


def mcweeny_purify_2rdm(rdm2_spin, conv=1.0e-07):
    """Perform 2-RDM purification based on McWeeny"s algorithm. The value
    selected for the convergence criteria should be consistent with the accuracy
    of the 2-RDM provided as input (in particular it should be correlated to the
    number of shots used to compute the RDM values, if a QPU or a shot-based
    quantum circuit simulator was used).

    This algorithm only works on a RDM associated with a 2-electron system.

    Args:
        rdm2_spin (numpy array): The 2-RDM to be purified (in chemistry
            notation).
        conv (float), optional: The convergence criteria for McWeeny"s
            purification.

    Returns:
        numpy.array: One-particle RDM in spatial orbital basis.
        numpy.array: Two-particle RDM in spatial orbital basis (in chemistry
            notation).
    """

    n_spinorbitals = rdm2_spin.shape[0]
    n_spatialorbitals = n_spinorbitals // 2
    rdm1_np = np.zeros((n_spatialorbitals, )*2)
    rdm2_np = np.zeros((n_spatialorbitals, )*4)

    # Transform the 2-RDM in physics notation
    D_matrix2 = np.asarray(rdm2_spin.transpose(0, 2, 1, 3), order="C")

    # Initialize trace of difference to 1.0, repeat McWeeny"s cycle until convergence is met (trace of D-D^{2} ~ 0.)
    diff2 = 1.0
    while abs(diff2) > conv:

        # Update the 2-RDM based on recursion relation
        D2 = np.einsum("pqrs,rsuv->pquv", D_matrix2, D_matrix2)
        D3 = np.einsum("pqrs,rsuv->pquv", D_matrix2, D2)
        D_matrix2 = 3.0 * D2 - 2.0 * D3

        # Compute trace of D-D^{2}
        DD_mat = np.einsum("pqrs,rsuv->pquv", D_matrix2, D_matrix2)
        D_diff = DD_mat - D_matrix2
        diff2 = sum(D_diff[i, j, i, j] for i, j in itertools.product(range(n_spinorbitals), repeat=2))

    # Transform the 2-RDM to chemistry notation
    D_matrix2_final = np.asarray(D_matrix2.transpose(0, 2, 1, 3), order="C")

    # Construct 1-RDM using 2-RDM
    rdm1_np_temp = np.zeros((n_spinorbitals, )*2)
    for i, j, k in itertools.product(range(n_spinorbitals), repeat=3):
        rdm1_np_temp[i, j] += D_matrix2_final[i, j, k, k]
    for i, j in itertools.product(range(n_spinorbitals), repeat=2):
        rdm1_np[i//2, j//2] += rdm1_np_temp[i, j]

    # Construct 2-RDM
    for i, j, k, l in itertools.product(range(n_spinorbitals), repeat=4):
        rdm2_np[i//2, j//2, k//2, l//2] += D_matrix2_final[i, j, k, l]

    return rdm1_np, rdm2_np



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import numpy as np

from tangelo.linq import Simulator
from tangelo.algorithms import BuiltInAnsatze, VQESolver
from tangelo.molecule_library import mol_H2_sto3g, mol_H4_sto3g, mol_H4_cation_sto3g, mol_NaH_sto3g, mol_H4_sto3g_symm
from tangelo.toolboxes.ansatz_generator.uccsd import UCCSD
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping
from tangelo.toolboxes.molecular_computation.rdms import matricize_2rdm
from tangelo.toolboxes.optimizers.rotosolve import rotosolve


class VQESolverTest(unittest.TestCase):

    def test_instantiation_vqe(self):
        """Try instantiating VQESolver with basic input."""

        options = {"molecule": mol_H2_sto3g, "qubit_mapping": "jw"}
        VQESolver(options)

    def test_instantiation_vqe_incorrect_keyword(self):
        """Instantiating with an incorrect keyword should return an error """

        options = {"molecule": mol_H2_sto3g, "qubit_mapping": "jw", "dummy": True}
        self.assertRaises(KeyError, VQESolver, options)

    def test_instantiation_vqe_missing_molecule(self):
        """Instantiating with no molecule should return an error."""

        options = {"qubit_mapping": "jw"}
        self.assertRaises(ValueError, VQESolver, options)

    def test_get_resources_h2_mappings(self):
        """Resource estimation, with UCCSD ansatz, given initial parameters.
        Each of JW, BK, and scBK mappings are checked.
        """
        mappings = ["jw", "bk", "scbk", "jkmn"]
        expected_values = [(15, 4), (15, 4), (5, 2), (15, 4)]

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw",
                        "initial_var_params": [0.1, 0.1]}
        for index, mi in enumerate(mappings):
            vqe_options["qubit_mapping"] = mi
            vqe_solver = VQESolver(vqe_options)
            vqe_solver.build()
            resources = vqe_solver.get_resources()

            self.assertEqual(resources["qubit_hamiltonian_terms"], expected_values[index][0])
            self.assertEqual(resources["circuit_width"], expected_values[index][1])

    def test_energy_estimation_vqe(self):
        """A single VQE energy evaluation for H2, using optimal parameters and
        exact simulator.
        """

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw"}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.energy_estimation([5.86665842e-06, 5.65317429e-02])
        self.assertAlmostEqual(energy, -1.137270422018, places=6)

    def test_operator_expectation_vqe(self):
        """ A test of the operator_expectation function, using optimal parameters and exact simulator """

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": 'jw'}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        # Test using var_params input and Qubit Hamiltonian
        energy = vqe_solver.operator_expectation(vqe_solver.qubit_hamiltonian, var_params=[5.86665842e-06, 5.65317429e-02])
        self.assertAlmostEqual(energy, -1.137270422018, places=6)

        # Test using updated var_params and Fermion Hamiltonian
        vqe_solver.ansatz.update_var_params([5.86665842e-06, 5.65317429e-02])
        energy = vqe_solver.operator_expectation(mol_H2_sto3g.fermionic_hamiltonian)
        self.assertAlmostEqual(energy, -1.137270422018, places=6)

        # Test the three in place operators
        n_electrons = vqe_solver.operator_expectation('N')
        self.assertAlmostEqual(n_electrons, 2, places=6)
        spin_z = vqe_solver.operator_expectation('Sz')
        self.assertAlmostEqual(spin_z, 0, places=6)
        spin2 = vqe_solver.operator_expectation('S^2')
        self.assertAlmostEqual(spin2, 0, places=6)

    def test_simulate_h2(self):
        """Run VQE on H2 molecule, with UCCSD ansatz, JW qubit mapping, initial
        parameters, exact simulator.
        """

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw",
                        "initial_var_params": [0.1, 0.1], "verbose": True}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.137270422018, delta=1e-4)

    def test_simulate_h2_with_deflation(self):
        """Run VQE on H2 molecule, with UCCSD ansatz, JW qubit mapping, initial
        parameters, exact simulator. Followed by UpCCGSD with deflation of ground state.
        """

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw",
                        "initial_var_params": [0.1, 0.1], "verbose": True}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.137270422018, delta=1e-4)

        # Use deflation to get first excited state.
        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UpCCGSD, "qubit_mapping": "jw",
                       "verbose": True, "deflation_circuits": [vqe_solver.optimal_circuit],
                       "deflation_coeff": 1.0, "ref_state": [0, 1, 0, 1]}
        vqe_solver_2 = VQESolver(vqe_options)
        vqe_solver_2.build()

        energy = vqe_solver_2.simulate()
        self.assertAlmostEqual(energy, -0.53247, delta=1e-4)

    def test_simulate_qmf_h2(self):
        """Run VQE on H2 molecule, with QMF ansatz, JW qubit mapping, initial
        parameters, exact simulator.
        """

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.QMF, "qubit_mapping": "jw",
                        "verbose": False}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.116684, delta=1e-4)

    def test_simulate_qcc_h2(self):
        """Run VQE on H2 molecule, with QCC ansatz, JW qubit mapping, initial
        parameters, exact simulator.
        """
        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.QCC, "qubit_mapping": "jw",
                        "verbose": False}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.137270, delta=1e-4)

    def test_simulate_ilc_h2(self):
        """Run VQE on H2 molecule, with ILC ansatz, JW qubit mapping, initial
        parameters, exact simulator.
        """

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.ILC, "qubit_mapping": "jw",
                        "verbose": False}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.137270, delta=1e-4)

    def test_simulate_vsqs_h2(self):
        """Run VQE on H2 molecule, with vsqs ansatz, JW qubit mapping, exact simulator for both molecule input and
        qubit_hamiltonian/hini/reference_state input
        """
        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.VSQS, "qubit_mapping": "jw",
                        "verbose": False, "ansatz_options": {"intervals": 3, "time": 3}}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.137270, delta=1e-4)

        qubit_hamiltonian = vqe_solver.qubit_hamiltonian
        h_init = vqe_solver.ansatz.h_init
        reference_state = vqe_solver.ansatz.prepare_reference_state()

        vqe_options = {"molecule": None, "qubit_hamiltonian": qubit_hamiltonian, "ansatz": BuiltInAnsatze.VSQS, "qubit_mapping": "jw",
                        "ansatz_options": {"intervals": 3, "time": 3, "qubit_hamiltonian": qubit_hamiltonian,
                                          "h_init": h_init, "reference_state": reference_state}}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.137270, delta=1e-4)

    def test_simulate_h2_qiskit(self):
        """Run VQE on H2 molecule, with UCCSD ansatz, JW qubit mapping, initial
        parameters, exact qiskit simulator.
        """

        backend_options = {"target": "qiskit", "n_shots": None, "noise_model": None}
        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw",
                        "initial_var_params": [6.28531447e-06, 5.65431626e-02], "verbose": True,
                        "backend_options": backend_options}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()
        energy = vqe_solver.simulate()

        self.assertAlmostEqual(energy, -1.13727042117, delta=1e-6)

    def test_simulate_h4(self):
        """Run VQE on H4 molecule, with UCCSD ansatz, JW qubit mapping, initial
        parameters, exact simulator.
        """
        vqe_options = {"molecule": mol_H4_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw",
                        "initial_var_params": "MP2", "verbose": False}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.9778312978826869, delta=1e-4)

    def test_simulate_h4_symm(self):
        """Run VQE on H2 molecule with symmetry turned on, with UCCSD ansatz, JW qubit mapping, initial
        parameters, exact simulator.
        """

        vqe_options = {"molecule": mol_H4_sto3g_symm, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw",
                       "initial_var_params": "MP2", "verbose": False}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.97783, delta=1e-4)

    def test_simulate_qmf_h4(self):
        """Run VQE on H4 molecule, with QMF ansatz, JW qubit mapping, initial
        parameters, exact simulator.
        """

        vqe_options = {"molecule": mol_H4_sto3g, "ansatz": BuiltInAnsatze.QMF, "qubit_mapping": "jw",
                        "verbose": False}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.789483, delta=1e-4)

    def test_simulate_qcc_h4(self):
        """Run VQE on H4 molecule, with QCC ansatz, JW qubit mapping, initial
        parameters, exact simulator.
        """

        vqe_options = {"molecule": mol_H4_sto3g, "ansatz": BuiltInAnsatze.QCC, "qubit_mapping": "jw",
                        "verbose": False}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.963270, delta=1e-4)

    def test_simulate_ilc_h4(self):
        """Run VQE on H4 molecule, with ILC ansatz, JW qubit mapping, initial
        parameters, exact simulator.
        """

        vqe_options = {"molecule": mol_H4_sto3g, "ansatz": BuiltInAnsatze.ILC, "qubit_mapping": "jw",
                        "verbose": False}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.960877, delta=1e-4)

    def test_simulate_h4_open(self):
        """Run VQE on H4 molecule, with UCCSD ansatz, JW qubit mapping, initial parameters, exact simulator """
        vqe_options = {"molecule": mol_H4_cation_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw",
                        "initial_var_params": "random", "verbose": False}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.6394, delta=1e-3)

    def test_simulate_qmf_h4_open(self):
        """Run VQE on H4 + molecule, with QMF ansatz, JW qubit mapping, initial
        parameters, exact simulator.
        """

        vqe_options = {"molecule": mol_H4_cation_sto3g, "ansatz": BuiltInAnsatze.QMF, "qubit_mapping": "jw",
                        "verbose": False}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.585918, delta=1e-4)

    def test_simulate_qcc_h4_open(self):
        """Run VQE on H4 + molecule, with QCC ansatz, JW qubit mapping, initial
        parameters, exact simulator.
        """

        vqe_options = {"molecule": mol_H4_cation_sto3g, "ansatz": BuiltInAnsatze.QCC, "qubit_mapping": "jw",
                        "verbose": False}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.638020, delta=1e-4)

    def test_simulate_ilc_h4_open(self):
        """Run VQE on H4 + molecule, with ILC ansatz, JW qubit mapping, initial
        parameters, exact simulator.
        """

        vqe_options = {"molecule": mol_H4_cation_sto3g, "ansatz": BuiltInAnsatze.ILC, "qubit_mapping": "jw",
                        "verbose": False}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.638020, delta=1e-4)

    def test_optimal_circuit_h4(self):
        """Run VQE on H4 molecule, save optimal circuit. Verify it yields
        optimal energy.
        """
        vqe_options = {"molecule": mol_H4_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw",
                        "initial_var_params": "MP2", "verbose": False}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()
        energy = vqe_solver.simulate()

        sim = Simulator()
        self.assertAlmostEqual(energy, sim.get_expectation_value(vqe_solver.qubit_hamiltonian, vqe_solver.optimal_circuit),
                                delta=1e-10)

    def test_get_rdm_h2(self):
        """Compute RDMs with UCCSD ansatz, JW qubit mapping, optimized
        parameters, exact simulator (H2).
        """

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw"}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        # Compute RDM matrices
        one_rdm, two_rdm = vqe_solver.get_rdm([5.86665842e-06, 5.65317429e-02])

        # Test traces of matrices
        n_elec, n_orb = mol_H2_sto3g.n_active_electrons, mol_H2_sto3g.n_active_mos
        self.assertAlmostEqual(np.trace(one_rdm), n_elec, msg="Trace of one_rdm does not match number of electrons",
                                delta=1e-6)
        rho = matricize_2rdm(two_rdm, n_orb)
        self.assertAlmostEqual(np.trace(rho), n_elec * (n_elec - 1),
                                msg="Trace of two_rdm does not match n_elec * (n_elec-1)", delta=1e-6)

    def test_get_rdm_h4(self):
        """Compute RDMs with UCCSD ansatz, JW qubit mapping, optimized
        parameters, exact simulator (H4).
        """

        vqe_options = {"molecule": mol_H4_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw"}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        # Compute RDM matrices
        var_params = [-6.47627367e-06, -5.24257363e-06, -5.99540594e-06, -7.70205325e-06, 1.15628926e-02,
                      3.42313563e-01,  3.48211343e-02,  1.49150233e-02, 7.53406401e-02,  8.44095525e-03,
                      -1.79981377e-01, -1.00585201e-01, 1.02162534e-02, -3.65870070e-02]
        one_rdm, two_rdm = vqe_solver.get_rdm(var_params)
        # Test traces of matrices
        n_elec, n_orb = mol_H4_sto3g.n_active_electrons, mol_H4_sto3g.n_active_mos
        self.assertAlmostEqual(np.trace(one_rdm), n_elec, msg="Trace of one_rdm does not match number of electrons",
                                delta=1e-6)
        rho = matricize_2rdm(two_rdm, n_orb)
        self.assertAlmostEqual(np.trace(rho), n_elec * (n_elec - 1),
                                msg="Trace of two_rdm does not match n_elec * (n_elec-1)", delta=1e-6)

    def test_custom_vqe(self):
        """VQE with custom optimizer and non-optimal variational parameters."""

        # Define and assign custom optimizer: cobyla
        def cobyla_oneshot_optimizer(func, var_params):
            from scipy.optimize import minimize
            result = minimize(func, var_params, method="COBYLA", options={"disp": True, "maxiter": 100})
            return result.fun, result.x

        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw",
                        "initial_var_params": "ones", "verbose": False,
                        "optimizer": cobyla_oneshot_optimizer}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.137270422018, places=6)

    def test_mapping_BK(self):
        """Test that BK mapping recovers the expected result, to within 1e-6 Ha,
        for the example of H2 and MP2 initial guess.
        """
        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "initial_var_params": "MP2", "verbose": False,
                        "qubit_mapping": "bk"}

        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()
        energy = vqe_solver.simulate()

        energy_target = -1.137270
        self.assertAlmostEqual(energy, energy_target, places=5)

    def test_mapping_scBK(self):
        """Test that scBK mapping recovers the expected result, to within
        1e-6 Ha, for the example of H2 and MP2 initial guess.
        """
        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "initial_var_params": "MP2", "verbose": False,
                        "qubit_mapping": "scbk"}

        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()
        energy = vqe_solver.simulate()

        energy_target = -1.137270
        self.assertAlmostEqual(energy, energy_target, places=5)

    def test_mapping_jkmn(self):
        """Test that JKMN mapping recovers the expected result, to within
        1e-6 Ha, for the example of H2 and MP2 initial guess.
        """
        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "initial_var_params": "MP2", "verbose": False,
                        "qubit_mapping": "jkmn"}

        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()
        energy = vqe_solver.simulate()

        energy_target = -1.137270
        self.assertAlmostEqual(energy, energy_target, places=5)

    def test_spin_reorder_equivalence(self):
        """Test that re-ordered spin input (all up followed by all down) returns
        the same optimized energy result for both JW and BK mappings.
        """
        vqe_options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD, "initial_var_params": "MP2", "up_then_down": True,
                        "verbose": False, "qubit_mapping": "jw"}

        vqe_solver_jw = VQESolver(vqe_options)
        vqe_solver_jw.build()
        energy_jw = vqe_solver_jw.simulate()

        vqe_options["qubit_mapping"] = "bk"
        vqe_solver_bk = VQESolver(vqe_options)
        vqe_solver_bk.build()
        energy_bk = vqe_solver_bk.simulate()

        energy_target = -1.137270
        self.assertAlmostEqual(energy_jw, energy_target, places=5)
        self.assertAlmostEqual(energy_bk, energy_target, places=5)

    def test_simulate_h4_frozen_orbitals(self):
        """Run VQE on H4 molecule, with UCCSD ansatz, JW qubit mapping, initial
        parameters, exact simulator. First (occupied) and last (virtual)
        orbitals are frozen.
        """
        mol_H4_sto3g_frozen = mol_H4_sto3g.freeze_mos([0, 3], inplace=False)

        vqe_options = {"molecule": mol_H4_sto3g_frozen, "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "jw",
                        "initial_var_params": "MP2", "verbose": False}
        vqe_solver = VQESolver(vqe_options)
        vqe_solver.build()

        energy = vqe_solver.simulate()
        self.assertAlmostEqual(energy, -1.8943598012229799, delta=1e-5)

    def test_simulate_nah_rucc(self):
        """Run VQE on NaH molecule, with UCC1 and UCC3 ansatze, JW qubit
        mapping. The computation is mapped to a HOMO-LUMO problem.
        """

        mol_NaH_sto3g_2mos = mol_NaH_sto3g.freeze_mos([i for i in range(9) if i not in [5, 9]], inplace=False)

        vqe_options = {"molecule": mol_NaH_sto3g_2mos, "ansatz": BuiltInAnsatze.UCC1, "qubit_mapping": "jw",
                        "initial_var_params": "zeros", "up_then_down": True, "verbose": False}

        vqe_solver_ucc1 = VQESolver(vqe_options)
        vqe_solver_ucc1.build()
        energy_ucc1 = vqe_solver_ucc1.simulate()

        vqe_options["ansatz"] = BuiltInAnsatze.UCC3
        vqe_solver_ucc3 = VQESolver(vqe_options)
        vqe_solver_ucc3.build()
        energy_ucc3 = vqe_solver_ucc3.simulate()

        self.assertAlmostEqual(energy_ucc1, -160.30334365109297, delta=1e-6)
        self.assertAlmostEqual(energy_ucc3, -160.30345935884606, delta=1e-6)

    def test_toomany_orbitals_rucc(self):
        """Test the case where there is too many orbitals in the system to be
        mapped into a HOMO-LUMO problem.
        """

        vqe_options = {"molecule": mol_NaH_sto3g, "ansatz": BuiltInAnsatze.UCC1, "qubit_mapping": "jw",
                        "initial_var_params": "zeros", "up_then_down": True, "verbose": False}

        with self.assertRaises(ValueError):
            vqe_solver_ucc1 = VQESolver(vqe_options)
            vqe_solver_ucc1.build()

        with self.assertRaises(ValueError):
            vqe_options["ansatz"] = BuiltInAnsatze.UCC3
            vqe_solver_ucc3 = VQESolver(vqe_options)
            vqe_solver_ucc3.build()

    def test_wrong_mapping_rucc(self):
        """Test the case where another mapping process (not JW) is selected."""

        mol_NaH_sto3g_2mos = mol_NaH_sto3g.freeze_mos([i for i in range(9) if i not in [5, 9]], inplace=False)

        vqe_options = {"molecule": mol_NaH_sto3g_2mos, "ansatz": BuiltInAnsatze.UCC1, "qubit_mapping": "bk",
                        "initial_var_params": "zeros", "up_then_down": True}

        with self.assertRaises(ValueError):
            vqe_solver_ucc1 = VQESolver(vqe_options)
            vqe_solver_ucc1.build()

        with self.assertRaises(ValueError):
            vqe_options["ansatz"] = BuiltInAnsatze.UCC3
            vqe_solver_ucc3 = VQESolver(vqe_options)
            vqe_solver_ucc3.build()

    def test_qubit_qhamiltonian_input(self):
        """Test the case where a qubit Hamiltonian is used to construct VQE."""

        qubit_hamiltonian = fermion_to_qubit_mapping(mol_H2_sto3g.fermionic_hamiltonian, mapping="jw")

        options = {"qubit_hamiltonian": qubit_hamiltonian,
                    "ansatz": UCCSD(mol_H2_sto3g, mapping="jw")}
        VQESolver(options)

    def test_qubit_qhamiltonian_input_conflicts(self):
        """Test the case where a molecule and a qubit Hamiltonian are passed as
        inputs.
        """

        qubit_hamiltonian = fermion_to_qubit_mapping(mol_H2_sto3g.fermionic_hamiltonian, mapping="jw")

        options = {"molecule": mol_H2_sto3g,
                    "qubit_hamiltonian": qubit_hamiltonian,
                    "ansatz": UCCSD(mol_H2_sto3g, mapping="jw")}

        with self.assertRaises(ValueError):
            VQESolver(options)

    def test_qubit_qhamiltonian_input_no_custom_ansatz(self):
        """Test the case where no custom ansatz is passed when using a qubit
        Hamiltonian as input.
        """

        qubit_hamiltonian = fermion_to_qubit_mapping(mol_H2_sto3g.fermionic_hamiltonian, mapping="jw")

        options = {"qubit_hamiltonian": qubit_hamiltonian}

        with self.assertRaises(TypeError):
            VQESolver(options).build()

        options["ansatz"] = BuiltInAnsatze.UCCSD

        with self.assertRaises(TypeError):
            VQESolver(options).build()

    def test_rotosolve(self):
        """Run VQE on H2 molecule, using Rotosolve, with UCC3 ansatz,
        JW qubit mapping, and exact simulator.
        """

        options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCC3,
                   "qubit_mapping": "jw", "verbose": False,
                   "optimizer": rotosolve, "up_then_down": True}

        vqe_solver = VQESolver(options)
        vqe_solver.build()
        vqe_solver.simulate()
        energy = vqe_solver.optimal_energy

        self.assertAlmostEqual(energy, -1.137270422018, delta=1e-4)

    def test_rotosolve_unsupported(self):
        """Test Rotosolve unsupported Ansatz with UCCSD on H2 molecule with
        JW qubit mapping, and exact simulator.
        """

        options = {"molecule": mol_H2_sto3g, "ansatz": BuiltInAnsatze.UCCSD,
                   "qubit_mapping": "jw", "optimizer": rotosolve}

        with self.assertRaises(ValueError):
            VQESolver(options).build()


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file provides an API enabling the use of adaptive classical shadows.
This algorithm is described in C. Hadfield, ArXiv:2105.12207 [Quant-Ph] (2021).
"""

from math import sqrt
import random

import numpy as np

from tangelo.toolboxes.measurements import ClassicalShadow
from tangelo.linq.circuit import Circuit
from tangelo.linq.helpers.circuits.measurement_basis import measurement_basis_gates, pauli_string_to_of


class AdaptiveClassicalShadow(ClassicalShadow):
    """Classical shadows using adaptive single Pauli measurements, as defined
    in C. Hadfield, ArXiv:2105.12207 [Quant-Ph] (2021).
    """

    def build(self, n_shots, qu_op):
        """Adaptive classical shadow building method to define relevant
        unitaries depending on the qubit operator.

        Args:
            n_shots (int): The number of desired measurements.
            qu_op (QubitOperator): The observable that one wishes to measure.

        Returns:
            list of str: The list of Pauli words that describes the measurement
                basis to use.
        """

        measurement_procedure = [self._choose_measurement(qu_op) for _ in range(n_shots)]

        self.unitaries += measurement_procedure
        return measurement_procedure

    def _choose_measurement(self, qu_op):
        """Algorithm 1 from the publication.

        Args:
            qu_op (QubitOperator): The operator that one wishes to maximize the
                measurement budget over.

        Returns:
            str: Pauli words for one measurement.
        """

        # Random bijection i: [n] -> [n]. Also, compute the inverse to undo it.
        i_qubit_random = random.sample(range(self.n_qubits), self.n_qubits)
        inverse_map = np.argsort(i_qubit_random)

        single_measurement = [None] * self.n_qubits

        # Choose measurement one qubit at the time.
        for it, i_qubit in enumerate(i_qubit_random):
            probs = self._get_probs(qu_op,
                                    i_qubit_random[0:it],
                                    single_measurement[0:it],
                                    i_qubit)

            single_measurement[it] = np.random.choice(["X", "Y", "Z"], size=None, replace=True, p=probs)

        # Reorder according to the qubit indices 0, 1, 2, ... self.n_qubits.
        reordered_measurement = [single_measurement[inverse_map[j]] for j in range(self.n_qubits)]

        return "".join(reordered_measurement)

    def _get_probs(self, qu_op, prev_qubits, prev_paulis, curr_qubit):
        """Generates the betas values from which the Pauli basis is determined
        for the current qubit (curr_qubit), as shown in Algorithm 2 from the
        paper.

        Args:
            qu_op (QubitOperator) : The operator one wishes to get the
                expectation value of.
            prev_qubits (list) : list of previous qubits from which the
                measurement basis is already determined.
            prev_paulis (list) : the Pauli word for prev_qubits.
            curr_qubit (int) : The current qubit being examined.

        Returns:
            list of float: cB values for X, Y and Z.
        """

        cbs = {"X": 0., "Y": 0., "Z": 0.}

        # Builds the candidate term (appending X, Y or Z). Then, transform
        # to a dictionary for removing qubit order dependency.
        B = dict(zip(prev_qubits, prev_paulis))

        for basis in cbs.keys():

            # Adds or overwrites the X, Y or Z prospect term.
            B[curr_qubit] = basis

            # Checks if term (P) is covered by candidate_pauli (B). P and B are
            # notation in the publication.
            for term, coeff in qu_op.terms.items():
                if not term:
                    continue

                # Like for B, remove qubit order dependency.
                P = dict(term)

                # Checks if an entry is in both dictionaries and compares the
                # values. If values are different, an entry is appended to
                # non_shared_items. If the key is not in P it is not. It means
                # that it is I for this qubit (so it does not break the cover
                # condition).
                non_shared_items = {k: B[k] for k in B if k in P and P[k] != B[k]}

                # If there are non-overlapping terms P_i not in {I, B_i(j)},
                # we do not take into account the term coefficient.
                if not non_shared_items:
                    cbs[basis] += coeff**2

        cbs = {basis: sqrt(cb) for basis, cb in cbs.items()}

        if sum(cbs.values()) < 1e-6:
            # Uniform distribution.
            probs = [1/3] * 3
        else:
            # Normalization + make sure there are in X, Y and Z order (eq. 3).
            sum_squared_cbs = sum([sqrt(cb) for cb in cbs.values()])
            probs = [sqrt(cbs[pauli]) / sum_squared_cbs for pauli in ["X", "Y", "Z"]]

        return probs

    def get_basis_circuits(self, only_unique=False):
        """Outputs a list of circuits corresponding to the adaptive single-Pauli
        unitaries.

        Args:
            only_unique (bool): Consider only unique unitaries.

        Returns:
            list of Circuit or tuple: All basis circuits or a tuple of unique
                circuits, string representation and numbers of occurence.
        """

        if not self.unitaries:
            raise ValueError(f"A set of unitaries must de defined (can be done with the build method in {self.__class__.__name__}).")

        unitaries_to_convert = self.unique_unitaries if only_unique else self.unitaries

        basis_circuits = list()
        for pauli_word in unitaries_to_convert:
            # Transformation of a unitary to quantum gates.
            pauli_of = pauli_string_to_of(pauli_word)
            basis_circuits += [Circuit(measurement_basis_gates(pauli_of), self.n_qubits)]

        # Counts each unique circuits (use for reversing to a full shadow from
        # an experiement on hardware).
        if only_unique:
            unique_basis_circuits = [(basis_circuits[i], u, self.unitaries.count(u)) for i, u in enumerate(unitaries_to_convert)]
            return unique_basis_circuits
        else:
            return basis_circuits

    def get_term_observable(self, term, coeff=1.):
        """Returns the estimated observable for a term and its coefficient.

        Args:
            term (tuple): Openfermion style of a qubit operator term.
            coeff (float): Multiplication factor for the term.

        Returns:
            float: Observable estimated with the shadow.
        """

        sum_product = 0
        n_match = 0

        # For every single_measurement in shadow_size.
        for snapshot in range(self.size):
            match = 1
            product = 1

            # Checks if there is a match for all Pauli gate in the term. Works
            # also with operator not on all qubits (e.g. X1 will hit Z0X1, Y0X1
            # and Z0X1).
            for i_qubit, pauli in term:
                if pauli != self.unitaries[snapshot][i_qubit]:
                    match = 0
                    break
                if self.bitstrings[snapshot][i_qubit] != "0":
                    product *= -1

            # No quantity is considered if there is no match.
            sum_product += match * product
            n_match += match

        return sum_product / n_match * coeff if n_match > 0 else 0.



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from openfermion.transforms import bravyi_kitaev as openfermion_bravyi_kitaev
from tangelo.toolboxes.qubit_mappings.bravyi_kitaev import bravyi_kitaev
from tangelo.toolboxes.operators import FermionOperator, QubitOperator


class BravyiKitaevTest(unittest.TestCase):

    def test_few_qubits(self):
        """Test that an error is raised if the number of qubits specified for an
        operator is too few.
        """
        # Instantiate simple non-trivial FermionOperator input
        input_operator = FermionOperator(((0, 0), (1, 0), (5, 0)))
        n_qubits = 3
        with self.assertRaises(ValueError):
            bravyi_kitaev(input_operator, n_qubits)

    def test_input_raise(self):
        """Test that invalid operator type throws an error."""
        input_operator = QubitOperator((1, 'X'))
        with self.assertRaises(TypeError):
            bravyi_kitaev(input_operator, n_qubits=2)

    def test_openfermion_equivalence(self):
        """Test that our wrapper returns the same result as openfermion's bare
        implementation of bravyi_kitaev.
        """
        # Instantiate simple non-trivial FermionOperator input
        input_operator = FermionOperator(((0, 0), (1, 0), (2, 0), (12, 1)))
        input_operator += FermionOperator((13, 1), 0.2)
        n_qubits = 14

        tangelo_result = bravyi_kitaev(input_operator, n_qubits=n_qubits)
        openfermion_result = openfermion_bravyi_kitaev(input_operator, n_qubits=n_qubits)

        # check that the number of terms is the same.
        self.assertEqual(len(tangelo_result.terms), len(openfermion_result.terms), msg="Number of terms generated does not agree"
                                                                                       "with openfermion implementation of Bravyi Kitaev.")

        # check that the term coefficients are the same
        for ti in tangelo_result.terms:
            factor = tangelo_result.terms[ti]
            openfermion_factor = openfermion_result.terms[ti]
            self.assertEqual(factor, openfermion_factor, msg="Term coefficient does not agree with openfermion bravyi_kitaev.")


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import scipy.optimize as sp


def diis(energies, coeffs):
    """
    DIIS extrapolation, originally developped by Pulay in
    Chemical Physics Letters 73, 393-398 (1980)

    Args:
        energies (array-like): Energy expectation values for amplified noise rates
        coeffs (array-like): Noise rate amplification factors

    Returns:
        float: Extrapolated energy
    """
    return extrapolation(energies, coeffs, 1)


def richardson(energies, coeffs, estimate_exp=False):
    """
    General, DIIS-like extrapolation procedure as found in
    Nature 567, 491-495 (2019) [arXiv:1805.04492]

    Args:
        energies (array-like): Energy expectation values for amplified noise rates
        coeffs (array-like): Noise rate amplification factors

    Returns:
        float: Extrapolated energy
    """
    if estimate_exp is False:
        # If no exponent estimation, run the direct Richardson solution
        return richardson_analytical(energies, coeffs)
    else:
        # For exponent estimation run the Richardson recursive algorithm
        return richardson_with_exp_estimation(energies, coeffs)


def extrapolation(energies, coeffs, taylor_order=None):
    """
    General, DIIS-like extrapolation procedure as found in
    Nature 567, 491-495 (2019) [arXiv:1805.04492]

    Args:
        energies (array-like): Energy expectation values for amplified noise rates
        coeffs (array-like): Noise rate amplification factors
        taylor_order (int): Taylor expansion order; None for Richardson extrapolation (order determined from number of datapoints), 1 for DIIS extrapolation

    Returns:
        float: Extrapolated energy
    """
    n = len(coeffs)
    if taylor_order is None:
        # Determine the expansion order in case of Richardson extrapolation
        taylor_order = n-1
    Eh = np.array(energies)
    coeffs = np.array(coeffs)

    # Setup the linear system matrix
    ck = np.array([coeffs**k for k in range(1, taylor_order+1)])
    B = np.ones((n+1, n+1))
    B[n, n] = 0
    B[:n, :n] = ck.T @ ck

    # Setup the free coefficients
    b = np.zeros(n+1)
    b[n] = 1  # For the Lagrange multiplier

    # Solve  the DIIS equations by least squares
    x = np.linalg.lstsq(B, b, rcond=None)[0]
    return np.dot(Eh, x[:-1])


def richardson_analytical(energies, coeffs):
    """
    Richardson extrapolation explicit result as found in
    Phys. Rev. Lett. 119, 180509 [arXiv:1612.02058]

    Args:
        energies (array-like): Energy expectation values for amplified noise rates
        coeffs (array-like): Noise rate amplification factors

    Returns:
        float: Extrapolated energy
    """
    Eh = np.array(energies)
    ck = np.array(coeffs)
    x = np.array([np.prod(ai/(a - ai)) for i, a in enumerate(ck)
                      for ai in [np.delete(ck, i)]])
    return np.dot(Eh, x)


def richardson_with_exp_estimation(energies, coeffs):
    """
    Richardson extrapolation by recurrence, with exponent estimation

    Args:
        energies (array-like): Energy expectation values for amplified noise rates
        coeffs (array-like): Noise rate amplification factors

    Returns:
        float: Extrapolated energy
    """
    n = len(coeffs)
    Eh = np.array(energies)
    c = np.array(coeffs)
    ck = np.array(coeffs)
    p, p_old = 1, 0

    # Define a helper function for exponent optimization
    def energy_diff(k, ti, si):
        tk = np.sign(ti)*np.abs(ti)**k
        sk = np.sign(si)*np.abs(si)**k
        Et = (tk*Eh[1] - Eh[0])/(tk - 1)
        Es = (sk*Eh[2] - Eh[0])/(sk - 1)
        return (Et - Es)**2

    # Run the Richardson algorithm with exponent optimization
    for i in range(n-1):
        ti = ck[0]/ck[1]
        si = ck[0]/ck[2]
        if ((n > 2) and (i < (n-2))):
            # Minimize the energy difference to determine the optimal exponent
            p = sp.minimize(energy_diff, p+1, args=(ti, si), method='BFGS', options={'disp': False}).x[0]
            if (i == 0):
                ck = c**p
        else:
            break

        # Run the main Richardson loop
        for j in range(n-i-1):
            ti = (ck[j]/ck[j+1])
            if (i > 0):
                dp = p - p_old
                if (np.isclose(dp, 0)):
                    break
                ck[j] = ck[j]*(c[j]**dp - c[j+1]**dp)/(ti - 1)
                ti = (ck[j]/ck[j+1])
            else:
                ck[j] = ck[j]*(c[j] - c[j+1])/(ti - 1)
            Eh[j] = (ti*Eh[j+1] - Eh[j])/(ti - 1)
        p_old = p
    return(Eh[0])



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import numpy as np

from tangelo.toolboxes.post_processing import diis, richardson

energies = [-1.04775574, -1.04302289, -1.03364568, -1.03005245]
coeffs = [1., 1.1, 1.2, 1.3]


class ExtrapolationTest(unittest.TestCase):

    def test_diis(self):
        """Test DIIS extrapolation on small sample data
        """
        calculated = diis(energies, coeffs)
        self.assertAlmostEqual(-1.11047933, calculated, delta=1e-6)

    def test_richardson(self):
        """Test Richardson extrapolation on small sample data
        """
        calculated = richardson(energies, coeffs)
        self.assertAlmostEqual(-1.45459036, calculated, delta=1e-6)
        calculated = richardson(energies, coeffs, estimate_exp=True)
        self.assertAlmostEqual(-1.05601603, calculated, delta=1e-6)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemsitry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from tangelo.algorithms.projective.quantum_imaginary_time import QITESolver
from tangelo.molecule_library import mol_H2_sto3g, mol_H4_sto3g
from tangelo.linq.noisy_simulation import NoiseModel


class QITESolverTest(unittest.TestCase):

    def test_instantiation_qite(self):
        """Try instantiating QITESolver with basic input."""

        options = {"molecule": mol_H2_sto3g, "qubit_mapping": "jw"}
        QITESolver(options)

    def test_instantiation_qite_incorrect_keyword(self):
        """Instantiating with an incorrect keyword should return an error """

        options = {"molecule": mol_H2_sto3g, "qubit_mapping": "jw", "dummy": True}
        self.assertRaises(KeyError, QITESolver, options)

    def test_instantiation_qite_missing_molecule(self):
        """Instantiating with no molecule should return an error."""

        options = {"qubit_mapping": "jw"}
        self.assertRaises(ValueError, QITESolver, options)

    def test_simulate_h2_noisy(self):
        """Run QITE on H2 molecule with bk qubit mapping and an empty noise model for 1 cycle.
        Result should be lower than mean field energy.
        """

        backend_options = {"target": None, "n_shots": 10000, "noise_model": NoiseModel()}

        qite_options = {"molecule": mol_H2_sto3g, "qubit_mapping": "scbk",
                        "verbose": True, "backend_options": backend_options,
                        "max_cycles": 1, "up_then_down": True}
        qite_solver = QITESolver(qite_options)
        qite_solver.build()

        energy = qite_solver.simulate()
        self.assertTrue(energy < mol_H2_sto3g.mf_energy)

    def test_simulate_h2(self):
        """Run QITE on H2 molecule, with JW qubit mapping and exact simulator
        """

        qite_options = {"molecule": mol_H2_sto3g, "qubit_mapping": "jw",
                        "verbose": False, "up_then_down": True}
        qite_solver = QITESolver(qite_options)
        qite_solver.build()

        energy = qite_solver.simulate()
        self.assertAlmostEqual(energy, -1.137270422018, delta=1e-4)

    def test_resources_h2(self):
        """Test get_resources funtion for QITE on H2 molecule, with JW qubit mapping.
        """

        qite_options = {"molecule": mol_H2_sto3g, "qubit_mapping": "jw",
                        "verbose": False}
        qite_solver = QITESolver(qite_options)
        qite_solver.build()
        resources = qite_solver.get_resources()
        self.assertEqual(resources["qubit_hamiltonian_terms"], 15)
        self.assertEqual(resources["pool_size"], 20)

    def test_mapping_BK(self):
        """Test that BK mapping recovers the expected result for the example of H2.
        """
        qite_options = {"molecule": mol_H2_sto3g, "verbose": False,
                        "qubit_mapping": "bk"}

        qite_solver = QITESolver(qite_options)
        qite_solver.build()
        energy = qite_solver.simulate()

        energy_target = -1.137270
        self.assertAlmostEqual(energy, energy_target, places=5)

    def test_mapping_JKMN(self):
        """Test that JKMN mapping recovers the expected result for the example of H2.
        """
        qite_options = {"molecule": mol_H2_sto3g, "verbose": False,
                        "qubit_mapping": "JKMN"}

        qite_solver = QITESolver(qite_options)
        qite_solver.build()
        energy = qite_solver.simulate()

        energy_target = -1.137270
        self.assertAlmostEqual(energy, energy_target, places=5)

    def test_mapping_scBK(self):
        """Test that scBK mapping recovers the expected result for the example of H2.
        """
        qite_options = {"molecule": mol_H2_sto3g, "verbose": False,
                        "qubit_mapping": "scbk", "up_then_down": True}

        qite_solver = QITESolver(qite_options)
        qite_solver.build()
        energy = qite_solver.simulate()

        energy_target = -1.137270
        self.assertAlmostEqual(energy, energy_target, places=5)

    def test_spin_reorder_equivalence(self):
        """Test that re-ordered spin input (all up followed by all down) returns
        the same optimized energy result for both JW and BK mappings.
        """
        qite_options = {"molecule": mol_H2_sto3g, "up_then_down": True,
                        "verbose": False, "qubit_mapping": "jw"}

        qite_solver_jw = QITESolver(qite_options)
        qite_solver_jw.build()
        energy_jw = qite_solver_jw.simulate()

        qite_options["qubit_mapping"] = "bk"
        qite_solver_bk = QITESolver(qite_options)
        qite_solver_bk.build()
        energy_bk = qite_solver_bk.simulate()

        energy_target = -1.137270
        self.assertAlmostEqual(energy_jw, energy_target, places=5)
        self.assertAlmostEqual(energy_bk, energy_target, places=5)

    def test_simulate_h4_frozen_orbitals(self):
        """Run QITE on H4 molecule, with UCCSD ansatz, JW qubit mapping, initial
        parameters, exact simulator. First (occupied) and last (virtual)
        orbitals are frozen.
        """
        mol_H4_sto3g_frozen = mol_H4_sto3g.freeze_mos([0, 3], inplace=False)

        qite_options = {"molecule": mol_H4_sto3g_frozen, "qubit_mapping": "jw",
                        "verbose": False}
        qite_solver = QITESolver(qite_options)
        qite_solver.build()

        energy = qite_solver.simulate()
        self.assertAlmostEqual(energy, -1.8943598012229799, delta=1e-5)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" This module defines a class that can be used to generate the circuit that
returns or uncomputes a given statevector (takes the given statevector to the zero state).
"""

import numpy as np
import math

from tangelo.linq import Circuit, Gate


class StateVector():
    """This class provides functions to either compute a statevector (of 2**n_qubits) from the zero state
    or take that state to the zero state

    Args:
        coefficients: The list or array of coefficients defining a state. Must have length 2**n_qubits where n_qubits is an integer.
    """

    def __init__(self, coefficients, order="msq_first"):
        n_qubits = math.log2(len(coefficients))
        if n_qubits == 0 or not n_qubits.is_integer():
            raise ValueError("Length of input state must be a power of 2")
        if order not in ["msq_first", "lsq_first"]:
            raise ValueError(f"order must be 'lsq_first' or 'msq_first'")

        self.n_qubits = int(n_qubits)
        self.coeffs = coefficients
        self.order = order

    def initializing_circuit(self, return_phase=False):
        """Calculate a circuit that implements the initialization of a given statevector from the zero state
        Implements a recursive initialization algorithm, including optimizations,
        from "Synthesis of Quantum Logic Circuits" Shende, Bullock, Markov
        https://arxiv.org/abs/quant-ph/0406176v5
        Additionally implements some extra optimizations: remove zero rotations and
        double cnots.

        Args:
            return_phase (bool): Return the global phase that is not captured by the circuit

        Returns:
            Circuit: The circuit that generates the statevector defined in coeffs
            float: If return_phase=True, the global phase angle not captured by the Circuit
        """
        # call to generate the circuit that takes the desired vector to zero
        disentangling_circuit, global_phase = self.uncomputing_circuit(return_phase=True)

        # invert the circuit to create the desired vector from zero (assuming
        # the qubits are in the zero state)
        state_prep_circuit = disentangling_circuit.inverse()
        global_phase = -global_phase

        return_value = (state_prep_circuit, global_phase) if return_phase else state_prep_circuit
        return return_value

    def uncomputing_circuit(self, return_phase=False):
        """Generate a circuit that takes the desired state to the zero state.

        Args:
            return_phase (bool): Flag to return global_phase

        Returns:
            Circuit: circuit to take self.coeffs vector to :math:`|{00\\ldots0}\\rangle`
            float: (if return_phase=True) The angle that defines the global phase not captured by the circuit
        """

        circuit = Circuit(n_qubits=self.n_qubits)

        # kick start the peeling loop, and disentangle one-by-one from LSB to MSB
        remaining_param = self.coeffs

        for i in range(self.n_qubits):
            # work out which rotations must be done to disentangle the LSB
            # qubit (we peel away one qubit at a time)
            remaining_param, thetas, phis = StateVector._rotations_to_disentangle(remaining_param)

            # perform the required rotations to decouple the LSB qubit (so that
            # it can be "factored" out, leaving a shorter amplitude vector to peel away)

            add_last_cnot = True
            if np.linalg.norm(phis) != 0 and np.linalg.norm(thetas) != 0:
                add_last_cnot = False

            if np.linalg.norm(phis) != 0:
                rz_mult = self._get_multiplex_circuit("RZ", phis, last_cnot=add_last_cnot)
                rz_mult.reindex_qubits(list(range(i, self.n_qubits)))
                circuit += rz_mult

            if np.linalg.norm(thetas) != 0:
                ry_mult = self._get_multiplex_circuit("RY", thetas, last_cnot=add_last_cnot)
                ry_mult.reindex_qubits(list(range(i, self.n_qubits)))
                for gate in reversed(ry_mult._gates):
                    circuit.add_gate(gate)
        global_phase = -np.angle(sum(remaining_param))

        if self.order == "lsq_first":
            circuit.reindex_qubits(list(reversed(range(0, self.n_qubits))))

        return_value = (circuit, global_phase) if return_phase else circuit
        return return_value

    @staticmethod
    def _rotations_to_disentangle(local_param):
        """Static internal method to work out Ry and Rz rotation angles used
        to disentangle the LSB qubit.
        These rotations make up the block diagonal matrix U (i.e. multiplexor)
        that disentangles the LSB.
        [[Ry(theta_1).Rz(phi_1)  0   .   .   0],
         [0         Ry(theta_2).Rz(phi_2) .  0],
                                    .
                                        .
          0         0           Ry(theta_2^n).Rz(phi_2^n)]]

        Args:
            local_param (array): The parameters of subset of qubits to return the LSB to the zero state.

        Returns:
            list of float: remaining vector with LSB set to |0>
            list of float: The necessary RY Gate parameters
            list of float: The necessary RZ Gate parameters
        """
        remaining_vector = []
        thetas = []
        phis = []

        param_len = len(local_param)

        for i in range(param_len // 2):
            # Ry and Rz rotations to move bloch vector from 0 to "imaginary"
            # qubit
            # (imagine a qubit state signified by the amplitudes at index 2*i
            # and 2*(i+1), corresponding to the select qubits of the
            # multiplexor being in state |i>)
            (remains, add_theta, add_phi) = StateVector._bloch_angles(
                local_param[2*i], local_param[2*i+1]
            )

            remaining_vector.append(remains)

            # rotations for all imaginary qubits of the full vector
            # to move from where it is to zero, hence the negative sign
            thetas.append(-add_theta)
            phis.append(-add_phi)

        return remaining_vector, thetas, phis

    @staticmethod
    def _bloch_angles(a_complex, b_complex):
        """Static internal method to work out rotation to create the passed-in
        qubit from the zero vector.

        Args:
            a_complex (complex): First complex number to calculate rotation from zero vector
            b_complex (complex): Second complex number to calculate rotation from zero vector

        Returns:
            complex: remaining phase angle not captured by RY and RZ
            float: calculated RY rotation angle
            float: calculated RZ rotation angle
        """

        # Force a and b to be complex, as otherwise numpy.angle might fail.
        a_complex = complex(a_complex)
        b_complex = complex(b_complex)
        mag_a = np.absolute(a_complex)
        final_r = float(np.sqrt(mag_a**2 + np.absolute(b_complex) ** 2))
        if final_r < np.finfo(float).eps:
            theta, phi, final_r, final_t = 0, 0, 0, 0
        else:
            theta = float(2 * np.arccos(mag_a / final_r))
            a_arg = np.angle(a_complex)
            b_arg = np.angle(b_complex)
            final_t = a_arg + b_arg
            phi = b_arg - a_arg

        return final_r * np.exp(1.0j * final_t / 2), theta, phi

    def _get_multiplex_circuit(self, target_gate, angles, last_cnot=True) -> Circuit:
        """Return a recursive implementation of a multiplexor circuit,
        where each instruction itself has a decomposition based on
        smaller multiplexors.
        The LSB is the multiplexor "data" and the other bits are multiplexor "select".

        Args:
            target_gate (Gate): Ry or Rz gate to apply to target qubit, multiplexed
                over all other "select" qubits
            angles (list[float]): list of rotation angles to apply Ry and Rz
            last_cnot (bool): add the last cnot if last_cnot = True

        Returns:
            Circuit: the circuit implementing the multiplexor's action
        """
        list_len = len(angles)
        local_n_qubits = int(math.log2(list_len)) + 1

        # case of no multiplexing: base case for recursion
        if local_n_qubits == 1:
            return Circuit([Gate(target_gate, 0, parameter=angles[0])], n_qubits=local_n_qubits)

        circuit = Circuit(n_qubits=local_n_qubits)

        lsb = 0
        msb = local_n_qubits - 1

        # calc angle weights, assuming recursion (that is the lower-level
        # requested angles have been correctly implemented by recursion
        angle_weight = np.kron([[0.5, 0.5], [0.5, -0.5]], np.identity(2 ** (local_n_qubits - 2)))

        # calc the combo angles
        angles = angle_weight.dot(np.array(angles)).tolist()

        # recursive step on half the angles fulfilling the above assumption
        multiplex_1 = self._get_multiplex_circuit(target_gate, angles[0: (list_len // 2)], False)
        circuit += multiplex_1

        # attach CNOT as follows, thereby flipping the LSB qubit
        circuit.add_gate(Gate('CNOT', target=lsb, control=msb))

        # implement extra efficiency from the paper of cancelling adjacent
        # CNOTs (by leaving out last CNOT and reversing (NOT inverting) the
        # second lower-level multiplex)
        multiplex_2 = self._get_multiplex_circuit(target_gate, angles[(list_len // 2):], False)
        if list_len > 1:
            for gate in reversed(multiplex_2._gates):
                circuit.add_gate(gate)

        else:
            circuit += multiplex_2

        # attach a final CNOT
        if last_cnot:
            circuit.add_gate(Gate("CNOT", lsb, msb))

        return circuit



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Define electronic structure solver employing the full configuration
interaction (CI) method.
"""

from pyscf import ao2mo, fci, mcscf

from tangelo.algorithms.electronic_structure_solver import ElectronicStructureSolver


class FCISolver(ElectronicStructureSolver):
    """ Uses the Full CI method to solve the electronic structure problem,
    through pyscf.

    Args:
        molecule (SecondQuantizedMolecule): The molecule to simulate.

    Attributes:
        ci (numpy.array): The CI wavefunction (float64).
        norb (int): The number of molecular orbitals.
        nelec (int): The number of electrons.
        cisolver (pyscf.fci.direct_spin0.FCI): The Full CI object.
        mean_field (pyscf.scf): Mean field object.
    """

    def __init__(self, molecule):

        self.ci = None
        self.norb = molecule.n_active_mos
        self.nelec = molecule.n_active_electrons
        self.spin = molecule.spin
        self.n_alpha = self.nelec//2 + self.spin//2 + (self.nelec % 2)
        self.n_beta = self.nelec//2 - self.spin//2

        # Need to use a CAS method if frozen orbitals are defined
        if molecule.frozen_mos is not None:
            # Generate CAS space with given frozen_mos, then use pyscf functionality to
            # obtain effective Hamiltonian with frozen orbitals excluded from the CI space.
            self.cas = True
            self.cassolver = mcscf.CASSCF(molecule.mean_field,
                                          molecule.n_active_mos,
                                          (self.n_alpha, self.n_beta))
            mos = self.cassolver.sort_mo([i+1 for i in molecule.active_mos])
            self.h1e_cas, self.ecore = self.cassolver.get_h1eff(mos)
            self.h2e_cas = self.cassolver.get_h2eff(mos)
            # Initialize the FCI solver that will use the effective Hamiltonian generated from CAS
            self.cisolver = fci.direct_spin1.FCI()
        else:
            self.cas = False
            if self.spin == 0:
                self.cisolver = fci.direct_spin0.FCI(molecule.to_pyscf(molecule.basis))
            else:
                self.cisolver = fci.direct_spin1.FCI()

        self.cisolver.verbose = 0
        self.mean_field = molecule.mean_field

    def simulate(self):
        """Perform the simulation (energy calculation) for the molecule.

        Returns:
            float: Total FCI energy.
        """

        if self.cas:  # Use previously generated effective Hamiltonian to obtain FCI solution
            energy, self.ci = self.cisolver.kernel(self.h1e_cas,
                                                   self.h2e_cas,
                                                   self.norb,
                                                   (self.n_alpha, self.n_beta),
                                                   ecore=self.ecore)
        else:  # Generate full Hamiltonian and obtain FCI solution.
            h1 = self.mean_field.mo_coeff.T @ self.mean_field.get_hcore() @ self.mean_field.mo_coeff

            twoint = self.mean_field._eri

            eri = ao2mo.restore(8, twoint, self.norb)
            eri = ao2mo.incore.full(eri, self.mean_field.mo_coeff)
            eri = ao2mo.restore(1, eri, self.norb)

            ecore = self.mean_field.energy_nuc()

            if self.spin == 0:
                energy, self.ci = self.cisolver.kernel(h1, eri, h1.shape[1], self.nelec, ecore=ecore)
            else:
                energy, self.ci = self.cisolver.kernel(h1, eri, h1.shape[1], (self.n_alpha, self.n_beta), ecore=ecore)

        return energy

    def get_rdm(self):
        """Compute the Full CI 1- and 2-particle reduced density matrices.

        Returns:
            numpy.array: One-particle RDM.
            numpy.array: Two-particle RDM.

        Raises:
            RuntimeError: If method "simulate" hasn't been run.
        """

        # Check if Full CI is performed
        if self.ci is None:
            raise RuntimeError("FCISolver: Cannot retrieve RDM. Please run the 'simulate' method first")

        if self.cas:
            one_rdm, two_rdm = self.cisolver.make_rdm12(self.ci, self.norb, (self.n_alpha, self.n_beta))
        else:
            if self.spin == 0:
                one_rdm = self.cisolver.make_rdm1(self.ci, self.norb, self.nelec)
                two_rdm = self.cisolver.make_rdm2(self.ci, self.norb, self.nelec)
            else:
                one_rdm, two_rdm = self.cisolver.make_rdm12(self.ci, self.norb, (self.n_alpha, self.n_beta))

        return one_rdm, two_rdm



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the ansatz abstract class, providing the foundation to
implement variational ansatz circuits.
"""

import abc


class Ansatz(abc.ABC):
    """Base class for all Ansatz. Derived/children classes wirtten by users and
    developers must implement the following abstract methods.
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def set_var_params(self):
        """Initialize variational parameters as zeros, random numbers, MP2, or
        any insightful values. Impacts the convergence of variational
        algorithms.
        """
        pass

    @abc.abstractmethod
    def prepare_reference_state(self):
        """Return circuit preparing the desired reference wavefunction (HF,
        multi-reference state...).
        """
        pass

    @abc.abstractmethod
    def build_circuit(self):
        """Build and return the quantum circuit implementing the ansatz."""
        pass

    @abc.abstractmethod
    def update_var_params(self):
        """Update value of variational parameters in the already-built ansatz
        circuit member.
        """
        pass



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the reduced UCCs ansatz class (RUCC, refering to both
UCC1 and UCC3), providing the foundation to implement variational ansatz
circuits. They are UCCD and UCCSD ansatz, but terms acting in the same way on an
Hartree-Fock initial state have been removed.

This must be used on a 2 levels system (2 MOs, 4 SOs) to be physically relevant.

Reference for those circuits.
    - McCaskey, A.J., Parks, Z.P., Jakowski, J. et al.
        Quantum chemistry as a benchmark for near-term quantum computers.
        npj Quantum Inf 5, 99 (2019).
        https://doi.org/10.1038/s41534-019-0209-0
"""

import numpy as np

from tangelo.linq import Circuit, Gate

from .ansatz import Ansatz


class RUCC(Ansatz):
    """This class implements the reduced-UCC ansatz, i.e. UCC1=UCCD and
    UCC3=UCCSD. Currently, only closed-shell is supported. This implies that the
    mean-field is computed with the RHF reference integrals.

    Args:
        n_var_params (int): Number of variational parameters, must be 1 or 3.
    """

    def __init__(self, n_var_params=1):

        if n_var_params not in {1, 3}:
            raise ValueError("The number of parameters for RUCC must be 1 or 3.")

        self.n_var_params = n_var_params

        # Supported reference state initialization.
        self.supported_reference_state = {"HF"}
        # Supported variational params initialization.
        self.supported_initial_var_params = {"zeros", "ones", "random"}

        # Default initial parameters for initialization.
        self.var_params_default = "zeros"
        self.reference_state_initialization = "HF"

        self.var_params = None
        self.circuit = None

    def set_var_params(self, var_params=None):
        """Set values for variational parameters, such as zeros, random numbers
        providing some keywords for users, and also supporting direct user input
        (list or numpy array). Return the parameters so that workflows such as
        VQE can retrieve these values.
        """

        if var_params is None:
            var_params = self.var_params_default

        if isinstance(var_params, str):
            if (var_params not in self.supported_initial_var_params):
                raise ValueError(f"Supported keywords for initializing variational parameters: {self.supported_initial_var_params}")
            if var_params == "ones":
                initial_var_params = np.ones((self.n_var_params,), dtype=float)
            elif var_params == "random":
                initial_var_params = np.random.random((self.n_var_params,))
            elif var_params == "zeros":
                initial_var_params = np.zeros((self.n_var_params,), dtype=float)
        else:
            initial_var_params = np.array(var_params)
            if initial_var_params.size != self.n_var_params:
                raise ValueError(f"Expected {self.n_var_params} variational parameters but "
                                 f"received {initial_var_params.size}.")
        self.var_params = initial_var_params
        return initial_var_params

    def prepare_reference_state(self):
        """Returns circuit preparing the reference state of the ansatz (e.g.
        prepare reference wavefunction with HF, multi-reference state, etc).
        This method outputs |1010>.

        Returns:
            Circuit: |1010> initial state.
        """

        if self.reference_state_initialization not in self.supported_reference_state:
            raise ValueError(f"Only supported reference state methods are:{self.supported_reference_state}")

        # NB: this one is consistent with JW but not other transforms.
        if self.reference_state_initialization == "HF":
            return Circuit([Gate("X", target=i) for i in (0, 2)])

    def build_circuit(self, var_params=None):
        """Build and return the quantum circuit implementing the state
        preparation ansatz (with currently specified initial_state and
        var_params).

        Args:
            list: Initial variational parameters. Must be consistent with the
                chosen UCC (1 or 3).
        """

        # Set initial variational parameters used to build the circuit.
        if var_params is not None:
            self.set_var_params(var_params)
        elif self.var_params is None:
            self.set_var_params()

        # Prepare reference state circuit |1010>.
        reference_state_circuit = self.prepare_reference_state()

        # If self.n_var_params has been changed manually, last check before building the circuit.
        if self.n_var_params == 1:
            rucc_circuit = self._ucc1()
        elif self.n_var_params == 3:
            rucc_circuit = self._ucc3()
        else:
            raise ValueError("The number of parameters for RUCC must be 1 or 3.")

        self.circuit = reference_state_circuit + rucc_circuit
        self.update_var_params(self.var_params)

    def update_var_params(self, var_params):
        """Shortcut: set value of variational parameters in the already-built
        ansatz circuit member. The circuit does not need to be rebuilt every
        time if only the variational parameters change.

        Args:
            list: Variational parameters to parse into the circuit.
        """

        assert len(var_params) == self.n_var_params
        self.var_params = var_params

        for param_index in range(self.n_var_params):
            self.circuit._variational_gates[param_index].parameter = var_params[param_index]

    def _ucc1(self):
        """This class implements the reduced-UCC ansatz UCC1. UCC1 is equivalent
        to the UCCD ansatz, but terms that act in the same manner of the HF
        state are removed.

        Returns:
            Circuit: UCC1 quantum circuit.
        """

        # Initialization of an empty list.
        lst_gates = list()

        # UCC1 gates are appended.
        lst_gates += [Gate("RX", 0, parameter=np.pi/2)]

        lst_gates += [Gate("H", qubit_i) for qubit_i in range(1, 4)]
        lst_gates += [Gate("CNOT", qubit_i+1, qubit_i) for qubit_i in range(3)]

        # Rz with var param to modifiy double excitation.
        lst_gates += [Gate("RZ", 3, parameter="theta", is_variational=True)]

        lst_gates += [Gate("CNOT", qubit_i, qubit_i-1) for qubit_i in range(3, 0, -1)]
        lst_gates += [Gate("H", qubit_i) for qubit_i in range(3, 0, -1)]

        lst_gates += [Gate("RX", 0, parameter=-np.pi/2)]

        return Circuit(lst_gates)

    def _ucc3(self):
        """This class implements the reduced-UCC ansatz UCC3. UCC3 is equivalent
        to the UCCSD ansatz, but terms that act in the same manner of the HF
        state are removed.

        Returns:
            Circuit: UCC3 quantum circuit.
        """

        # Initialization of an empty list.
        lst_gates = list()

        # UCC3 gates are appended.
        lst_gates += [Gate("RX", 0, parameter=np.pi/2)]
        lst_gates += [Gate("H", 1)]
        lst_gates += [Gate("RX", 2, parameter=np.pi/2)]
        lst_gates += [Gate("H", 3)]

        lst_gates += [Gate("CNOT", qubit_i, qubit_i-1) for qubit_i in (1, 3)]

        # Rz with var param to modifiy single excitations.
        lst_gates += [Gate("RZ", 1, parameter="theta0", is_variational=True)]
        lst_gates += [Gate("RZ", 3, parameter="theta1", is_variational=True)]

        lst_gates += [Gate("CNOT", 3, 2)]

        lst_gates += [Gate("RX", 2, parameter=-np.pi/2)]
        lst_gates += [Gate("H", 2)]

        lst_gates += [Gate("CNOT", qubit_i, qubit_i-1) for qubit_i in (2, 3)]

        # Rz with var param to modifiy double excitation.
        lst_gates += [Gate("RZ", 3, parameter="theta2", is_variational=True)]

        lst_gates += [Gate("CNOT", qubit_i, qubit_i-1) for qubit_i in range(3, 0, -1)]
        lst_gates += [Gate("H", qubit_i) for qubit_i in range(3, 0, -1)]

        lst_gates += [Gate("RX", 0, parameter=-np.pi/2)]

        return Circuit(lst_gates)



""" Abstract parent class encapsulating basic features of compute services for quantum circuit simulation """

import abc


class QpuConnection(abc.ABC):
    """ Abstract class encapsulating login/authentication setup and job submission and management  """

    @abc.abstractmethod
    def __init__(self):
        pass

    @abc.abstractmethod
    def job_submit(self):
        """ Submit a job to the compute services """
        pass

    @abc.abstractmethod
    def job_status(self, job_id):
        """ Retrieve information about a previously submitted job, through its job id """
        pass

    @abc.abstractmethod
    def job_results(self, job_id):
        """ Retrieve the results of previously submitted job, through its job id """
        pass

    @abc.abstractmethod
    def job_cancel(self, job_id):
        """ Attempt to cancel a previously submitted job, through its job id """
        pass
